/**
 * \file gre/lrlidar.h
 * \brief  GRE::LAYER_LIDAR class definition.
 *
 * \if NODOC
 * $Id: lrlidar.h_v 1.34 2012/10/15 13:49:37 mju Exp $
 *
 * $Log: lrlidar.h_v $
 * Revision 1.34  2012/10/15 13:49:37  mju
 * Remove isClassUsed.
 *
 * Revision 1.33  2012/10/12 21:29:56  mju
 * Add values for select/style/terrain using previous classification.
 *
 * Revision 1.32  2012/10/11 20:38:41  mju
 * Add terrainparms::hidepoints enum so can elect to not hide at all.
 *
 * Revision 1.31  2012/10/08 20:44:27  mju
 * Add members for using previous classification.
 *
 * Revision 1.30  2012/06/14 21:34:40  mju
 * Change class selection members to allow 256 classes as in LAS 1.4
 *
 * Revision 1.29  2011/11/21 16:42:52  mju
 * Add v_estimateRangeZ.
 *
 * Revision 1.28  2011/10/20 13:05:15  mju
 * Remove redundant hdrs.
 *
 * Revision 1.27  2011/09/07 21:26:35  mju
 * Add terrainparms hideAllPoints member.
 *
 * Revision 1.26  2011/09/07 14:13:05  mju
 * Change returnselection_All to include "other".
 *
 * Revision 1.25  2011/08/15 17:34:41  mju
 * Add v_dataTipGetString and predefined datatip values.
 *
 * Revision 1.24  2011/08/08 17:56:44  mju
 * Add isClassSelected/All for point/terrain.
 *
 * Revision 1.23  2011/08/05 22:14:34  mju
 * Add virtual methods for dense layer, default scale, etc.
 *
 * Revision 1.22  2011/08/05 17:58:57  mju
 * Various dataset adjustments.
 *
 * Revision 1.21  2011/07/20 13:48:10  mju
 * Begin supporting las dataset.
 *
 * Revision 1.20  2011/06/20 19:37:59  mju
 * Make point/terrainparms members private.
 *
 * Revision 1.19  2011/04/13 16:30:06  mju
 * Allow for returntype_NumTypes and all return numbers 0-7.
 *
 * Revision 1.18  2010/09/09 21:22:20  mju
 * Add terrainparms::shadeSmoothMethod.
 *
 * Revision 1.17  2010/09/02 17:38:29  mju
 * Use colorspread class.
 *
 * Revision 1.15  2010/07/23 16:19:56  mju
 * Add terrainparms.setDefaults.
 *
 * Revision 1.13  2010/07/08 21:08:20  mju
 * Add disppparm::terrainparms.
 *
 * Revision 1.12  2010/06/25 20:39:00  mju
 * Add getSpecialColorAndSize.
 *
 * Revision 1.11  2010/06/18 20:21:33  mju
 * Add pointparms::setDefaultClassColors.
 *
 * Revision 1.8  2010/06/11 21:33:31  mju
 * Add pointparm values for color spread.
 *
 * Revision 1.6  2010/06/03 21:36:33  mju
 * Add pointparms stuff for built-in (special) styling.
 *
 * Revision 1.5  2010/05/27 21:37:12  mju
 * Add getDbTableNumMain/Class.
 *
 * Revision 1.3  2010/05/12 21:04:10  mju
 * Add stylemode_ByReturn and change now returnselection defined.
 *
 * Revision 1.1  2010/04/23 16:54:33  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_GRE_LRLIDAR_H
#define	INC_GRE_LRLIDAR_H

#ifndef	INC_GRE_LAYER_H
#include <gre/layer.h>
#endif

#ifndef	INC_GRE_LEGEND_H
#include <gre/legend.h>
#endif

#ifndef	INC_GRE_ELEMPARMS_H
#include <gre/elemparms.h>
#endif

#ifndef	INC_MI32_SCRNUNIT_H
#include <mi32/scrnunit.h>
#endif

#ifndef	INC_MI32_COLORSPREAD_H
#include <mi32/colorspread.h>
#endif

#ifndef	INC_MI32_LIDARENUMS_H
#include <mi32/lidarenums.h>
#endif

#ifndef	INC_MI32_DATETIME_H
#include <mi32/datetime.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations
class STAT_UNIVARIATE;
namespace RVC {
	class SHAPE;
	class DBASE_SHAPE;
	}
namespace SPATMOD {
	namespace IMAGE {
		class PIXELTABLE;
		}
	}
namespace LIDAR {
	namespace LAS {
		class DATASET;
		class POINTRECORD;
		}
	}
#endif


namespace GRE {	
//===================================================================================================================

//! GRE layer class for viewing LIDAR objects.
class LAYER_LIDAR : public LAYER {
	public:

		#ifndef GENERATING_DOXYGEN_OUTPUT
		class PRIV;
		#endif

		enum COMPONENT {
			COMPONENT_Point = 0,
//			COMPONENT_Pulse,
			COMPONENT_COUNT
			};

		//! Lidar layer display parameters.
		class DISPPARM : public LAYER::DISPPARM {
			public:

				//! Which returns are selected.
				enum RETURNSELECTION {
					RETURNSELECTION_None =					0,
					RETURNSELECTION_Single =				0x01,		//!< Select single-return points
					RETURNSELECTION_FirstOfMultiple =	0x02,		//!< Select first of multiple returns
					RETURNSELECTION_LastOfMultiple =		0x04,		//!< Select last of multiple returns
					RETURNSELECTION_Intermediate =		0x08,		//!< Select intermediate returns
					RETURNSELECTION_Other =					0x10,		//!< Select other returns
					RETURNSELECTION_All =					0x1F
					};

				//! Class selection mode.
				enum CLASSSELECTMODE {
					CLASSSELECTMODE_Current =				0x00,		//!< Select based on current classification
					CLASSSELECTMODE_All =					0x01,		//!< Select all classes
					CLASSSELECTMODE_Previous =				0x02,		//!< Select based on previous classification
					CLASSSELECTMODE_Changed =				0x04,		//!< Select only if changed from previous, cannot combine with _Unchanged
					CLASSSELECTMODE_Unchanged =			0x08,		//!< Select only if unchanged from previous, cannot combine with _Changed
					CLASSSELECTMODE_Mask_Previous =		(CLASSSELECTMODE_Previous | CLASSSELECTMODE_Changed | CLASSSELECTMODE_Unchanged),
					};

				class DLG;										//!< Layer settings dialog, defined in <gre/uilidar.h>

				class POINTPARMS : public GRE::POINTPARMS {
					public:

						typedef GRE::POINTPARMS BASECLASS;

						enum ATTRIBUTE {
							ATTRIBUTE_None =					0,
							ATTRIBUTE_Class =					1,
							ATTRIBUTE_ReturnType =			2,
							ATTRIBUTE_ReturnNumber =		3,
							ATTRIBUTE_Intensity =			4,
							ATTRIBUTE_ScanAngle =			5,
							ATTRIBUTE_Elevation =			6,
							ATTRIBUTE_PrevClass =			7,
							ATTRIBUTE_PrevClassChange =	8,
							};

						enum DISTRIBUTION {
							DISTRIBUTION_Linear =		0,
							DISTRIBUTION_Equalize =		1,
							DISTRIBUTION_Normalize =	2,
							DISTRIBUTION_Logarithmic =	3,
							};

						//! Special datatip field values, set in DATATIPPARMS::m_Predefined.
						enum DATATIP {
							DATATIP_ClassName =				1,
							DATATIP_ReturnNumberCount =	2,
							};

						static const SERIALIZERITEM * SerialGetItemDef ();

						POINTPARMS ();
						POINTPARMS (const POINTPARMS& rhs);
						~POINTPARMS ();

						POINTPARMS& operator= (const POINTPARMS& rhs);

						UINT32* AccessClassSelection ()
							{ return (m_ClassSelection); }

						ERRVALUE ComputeSpreadColors (COLOR_ARRAY& SpreadColors, int NumColors) const;
						ERRVALUE ComputeSpreadColorsIndexed (COLOR_ARRAY& SpreadColors, SIMPLE_ARRAY<UINT8>& ColorIdxTable, const STAT_UNIVARIATE& stats) const;

						bool GetCheckIntensity () const
							{ return (m_CheckIntensity); }
						bool GetCheckScanAngle () const
							{ return (m_CheckScanAngle); }
						const COLOR& GetClassColor (UINT8 c) const
							{ return (m_ClassColor[c]); }
						const COLOR& GetClassPrevSameColor (UINT8 c) const
							{ return (m_ClassPrevSameColor[c]); }
						const UINT32* GetClassSelection () const
							{ return (m_ClassSelection); }
						CLASSSELECTMODE GetClassSelectMode () const
							{ return (m_ClassSelectMode); }
						ATTRIBUTE GetColorAttribute () const
							{ return (m_ColorAttribute); }
						DISTRIBUTION GetColorDistribution () const
							{ return (m_ColorDistribution); }
						const COLORSPREAD& GetColorSpread () const
							{ return (m_ColorSpread); }
						UINT16 GetColorSpreadLegendLength () const
							{ return (m_ColorSpreadLegendLength); }
						UINT16 GetIntensityMax () const
							{ return (m_IntensityMax); }
						UINT16 GetIntensityMin () const
							{ return (m_IntensityMin); }
						float GetNominalSpacing () const
							{ return (m_NominalSpacing); }
						const COLOR& GetReturnNumColor (UINT8 ReturnNum) const
							{ return (m_ReturnNumColor[ReturnNum]); }
						RETURNSELECTION GetReturnSelection () const
							{ return (m_ReturnSelection); }
						const COLOR& GetReturnTypeColor (UINT8 ReturnType) const
							{ return (m_ReturnTypeColor[ReturnType]); }
						INT8 GetScanAngleMax () const
							{ return (m_ScanAngleMax); }
						INT8 GetScanAngleMin () const
							{ return (m_ScanAngleMin); }
						bool GetShowDeleted () const
							{ return (m_ShowDeleted); }
						ATTRIBUTE GetSizeAttribute () const
							{ return (m_SizeAttribute); }
						float GetSizeDrawMax () const
							{ return (m_SizeDrawMax); }
						float GetSizeDrawMin () const
							{ return (m_SizeDrawMin); }
						float GetSizeSpreadPctMax () const
							{ return (m_SizeSpreadPctMax); }
						float GetSizeSpreadPctMin () const
							{ return (m_SizeSpreadPctMin); }
						SCREENUNIT GetSizeUnitAllSame () const
							{ return (m_SizeUnitAllSame); }
						SCREENUNIT GetSizeUnitRange () const
							{ return (m_SizeUnitRange); }

						bool IsClassSelected (UINT8 ClassNumber) const
							{ return ((m_ClassSelection[ClassNumber/32] & (1 << (ClassNumber % 32))) != 0); }
						bool IsClassSelectedAll () const
							{ return ((m_ClassSelectMode & CLASSSELECTMODE_All) != 0); }
						bool IsPreviousClassificationUsed () const;

						void SetCheckIntensity (bool CheckIntensity)
							{ m_CheckIntensity = CheckIntensity; }
						void SetCheckScanAngle (bool CheckScanAngle)
							{ m_CheckScanAngle = CheckScanAngle; }
						void SetClassColor (UINT8 c, const COLOR& color)
							{ m_ClassColor[c] = color; }
						void SetClassPrevSameColor (UINT8 c, const COLOR& color)
							{ m_ClassPrevSameColor[c] = color; }
						void SetClassSelectAll (bool SelectAll);
						void SetClassSelected (UINT8 ClassNumber, bool select)
							{ if (select) m_ClassSelection[ClassNumber/32] |= (1 << (ClassNumber % 32)); else m_ClassSelection[ClassNumber/32] &= ~(1 << (ClassNumber % 32)); }
						void SetClassSelectMode (CLASSSELECTMODE SelectMode)
							{ m_ClassSelectMode = SelectMode; }
						void SetColorAttribute (ATTRIBUTE ColorAttribute)
							{ m_ColorAttribute = ColorAttribute; }
						void SetColorDistribution (DISTRIBUTION ColorDistribution)
							{ m_ColorDistribution = ColorDistribution; }
						void SetColorSpread (const COLORSPREAD& ColorSpread)
							{ m_ColorSpread = ColorSpread; }
						void SetColorSpreadLegendLength (UINT16 ColorSpreadLegendLength)
							{ m_ColorSpreadLegendLength = ColorSpreadLegendLength; }

						void SetDefaults (RVC::DBASE_SHAPE& DbShape);

						void SetIntensityRange (UINT16 IntensityMin, UINT16 IntensityMax)
							{ m_IntensityMin = IntensityMin; m_IntensityMax = IntensityMax; }
						void SetNominalSpacing (float NominalSpacing)
							{ m_NominalSpacing = NominalSpacing; }
						void SetReturnSelection (RETURNSELECTION ReturnSelection)
							{ m_ReturnSelection = ReturnSelection; }
						void SetReturnNumColor (UINT8 ReturnNum, const COLOR& color)
							{ m_ReturnNumColor[ReturnNum] = color; }
						void SetReturnTypeColor (UINT8 ReturnType, const COLOR& color)
							{ m_ReturnTypeColor[ReturnType] = color; }
						void SetScanAngleRange (INT8 ScanAngleMin, INT8 ScanAngleMax)
							{ m_ScanAngleMin = ScanAngleMin; m_ScanAngleMax = ScanAngleMax; }
						void SetShowDeleted (bool ShowDeleted)
							{ m_ShowDeleted = ShowDeleted; }
						void SetSizeAttribute (ATTRIBUTE SizeAttribute)
							{ m_SizeAttribute = SizeAttribute; }
						void SetSizeDrawRange (float SizeDrawMin, float SizeDrawMax)
							{ m_SizeDrawMin = SizeDrawMin; m_SizeDrawMax = SizeDrawMax; }
						void SetSizeSpreadPctRange (float SizeSpreadPctMin, float SizeSpreadPctMax)
							{ m_SizeSpreadPctMin = SizeSpreadPctMin; m_SizeSpreadPctMax = SizeSpreadPctMax; }
						void SetSizeUnitAllSame (SCREENUNIT SizeUnitAllSame)
							{ m_SizeUnitAllSame = SizeUnitAllSame; }
						void SetSizeUnitRange (SCREENUNIT SizeUnitRange)
							{ m_SizeUnitRange = SizeUnitRange; }

						void Validate (RVC::DBASE_SHAPE& DbShape);

					private:
						#ifndef GENERATING_DOXYGEN_OUTPUT
						typedef POINTPARMS THISCLASS;

						UINT32 m_ClassSelection[8];				//!< Bit mask for which classes are selected to draw
						UINT16 m_IntensityMin;
						UINT16 m_IntensityMax;
						RETURNSELECTION m_ReturnSelection;		//!< Which returns are selected to draw
						INT8 m_ScanAngleMin;
						INT8 m_ScanAngleMax;
						bool m_CheckIntensity;
						bool m_CheckScanAngle;
						bool m_ShowDeleted;							//!< Show deleted points also
						CLASSSELECTMODE m_ClassSelectMode;
						ATTRIBUTE m_ColorAttribute;
						DISTRIBUTION m_ColorDistribution;
						COLORSPREAD m_ColorSpread;
						UINT16 m_ColorSpreadLegendLength;
						COLOR m_ClassColor[256];
						COLOR m_ReturnTypeColor[LIDAR::POINTRETURNTYPE_NUMTYPES];
						COLOR m_ReturnNumColor[8];
						COLOR m_ClassPrevSameColor[2];
						ATTRIBUTE m_SizeAttribute;
						DISTRIBUTION m_SizeDistribution;
						SCREENUNIT m_SizeUnitAllSame;
						SCREENUNIT m_SizeUnitRange;
						float m_SizeSpreadPctMin;
						float m_SizeSpreadPctMax;
						float m_SizeDrawMin;
						float m_SizeDrawMax;
						float m_NominalSpacing;

						void Copy (const POINTPARMS& rhs);
						void SetDefaultClassColors (RVC::DBASE_SHAPE& DbShape);
						#endif
					};

				class TERRAINPARMS {
					public:

						enum DISPLAYMODE {
							DISPLAYMODE_None =				0,
							DISPLAYMODE_ShadedRelief =		1,
							DISPLAYMODE_ColorElevation =	2,
							DISPLAYMODE_ColorShaded =		3,
							};

						enum COMPUTEMODE {
							COMPUTEMODE_Dynamic =	1,
							COMPUTEMODE_Static =		2
							};

						enum HIDEPOINTS {
							HIDEPOINTS_CloserThanMin =	0,
							HIDEPOINTS_All =				1,
							HIDEPOINTS_None =				2
							};

						static const SERIALIZERITEM * SerialGetItemDef ();

						TERRAINPARMS ();
						~TERRAINPARMS ();

						UINT32* AccessClassSelection ()
							{ return (m_ClassSelection); }

						ERRVALUE ComputeSpreadColors (SPATMOD::IMAGE::PIXELTABLE& PixelTable) const;

						const UINT32* GetClassSelection () const
							{ return (m_ClassSelection); }
						CLASSSELECTMODE GetClassSelectMode () const
							{ return (m_ClassSelectMode); }
						const COLORSPREAD& GetColorSpread () const
							{ return (m_ColorSpread); }
						DISPLAYMODE GetDisplayMode () const
							{ return (m_DisplayMode); }
						UINT8 GetElevDistribMethod () const
							{ return (m_ElevDistribMethod); }
						const DOUBLE_RANGE& GetElevRangePct () const
							{ return (m_ElevRangePct); }
						HIDEPOINTS GetHidePoints () const
							{ return (static_cast<HIDEPOINTS>(m_HidePoints)); }
						float GetMinPointSpacing () const
							{ return (m_MinPointSpacing); }
						RETURNSELECTION GetReturnSelection () const
							{ return (m_ReturnSelection); }
						double GetShadeAzimAngle () const
							{ return (m_ShadeAzimAngle); }
						double GetShadeElevAngle () const
							{ return (m_ShadeElevAngle); }
						UINT8 GetShadeSmoothMethod () const
							{ return (m_ShadeSmoothMethod); }
						double GetShadeZScale () const
							{ return (m_ShadeZScale); }

						bool IsClassSelected (UINT8 ClassNumber) const
							{ return ((m_ClassSelection[ClassNumber/32] & (1 << (ClassNumber % 32))) != 0); }
						bool IsClassSelectedAll () const
							{ return ((m_ClassSelectMode & CLASSSELECTMODE_All) != 0); }
						bool IsPreviousClassificationUsed () const;

						void SetClassSelectAll (bool SelectAll);
						void SetClassSelected (UINT8 ClassNumber, bool select)
							{ if (select) m_ClassSelection[ClassNumber/32] |= (1 << (ClassNumber % 32)); else m_ClassSelection[ClassNumber/32] &= ~(1 << (ClassNumber % 32)); }
						void SetClassSelectMode (CLASSSELECTMODE SelectMode)
							{ m_ClassSelectMode = SelectMode; }

						void SetColorSpread (const COLORSPREAD& ColorSpread)
							{ m_ColorSpread = ColorSpread; }

						void SetDefaults ();

						void SetDisplayMode (DISPLAYMODE DisplayMode)
							{ m_DisplayMode = DisplayMode; }

						//! Set elevation distribution method.
						void SetElevDistribMethod (
							UINT8 ElevDistribMethod									//!< CONTRASTPARM::METHOD
							) { m_ElevDistribMethod = ElevDistribMethod; }

						void SetElevRangePct (const DOUBLE_RANGE& ElevRangePct)
							{ m_ElevRangePct = ElevRangePct; }

						void SetHidePoints (HIDEPOINTS HidePoints)
							{ m_HidePoints = static_cast<UINT8>(HidePoints); }

						void SetMinPointSpacing (float MinPointSpacing)
							{ m_MinPointSpacing = MinPointSpacing; }

						void SetReturnSelection (RETURNSELECTION ReturnSelection)
							{ m_ReturnSelection = ReturnSelection; }

						void SetShadeAzimAngle (double ShadeAzimAngle)
							{ m_ShadeAzimAngle = ShadeAzimAngle; }

						void SetShadeElevAngle (double ShadeElevAngle)
							{ m_ShadeElevAngle = ShadeElevAngle; }

						void SetShadeSmoothMethod (UINT8 ShadeSmoothMethod)
							{ m_ShadeSmoothMethod = ShadeSmoothMethod; }

						void SetShadeZScale (double ShadeZScale)
							{ m_ShadeZScale = ShadeZScale; }

						void Validate ();

					private:
						#ifndef GENERATING_DOXYGEN_OUTPUT
						typedef TERRAINPARMS THISCLASS;

						DISPLAYMODE m_DisplayMode;
						UINT32 m_ClassSelection[8];				// Bit mask for which classes are selected to draw
						RETURNSELECTION m_ReturnSelection;		// Which returns are selected to draw
						double m_ShadeAzimAngle;
						double m_ShadeElevAngle;
						double m_ShadeZScale;
						DOUBLE_RANGE m_ElevRangePct;
						UINT8 m_ElevDistribMethod;
						UINT8 m_ShadeSmoothMethod;
						UINT8 m_HidePoints;
						CLASSSELECTMODE m_ClassSelectMode;
						COLORSPREAD m_ColorSpread;
						float m_MinPointSpacing;
						float m_MaxDetail;

						#endif
					};

				DISPPARM ();

				DISPPARM (
					const DISPPARM& rhs
					) { Copy(rhs); }

				virtual ~DISPPARM (
					) { Free(); }

				//! Assignment.
				DISPPARM& operator= (
					const DISPPARM& rhs
					) { if (this != &rhs) { Free(); Copy(rhs); } return (*this); }

				//! Get name of georeference subobject used.
				const RVC::OBJECTNAME& GetGeorefName () const
					{ return (m_GeorefName); }

				//! Get object item.
				const RVC::OBJITEM& GetObjItem () const
					{ return (m_ObjItem); }

				//! Get object item for style.
				const RVC::OBJITEM& GetObjItemStyle () const
					{ return (m_ObjItemStyle); }

				//! Get point parameters.
				const POINTPARMS& GetPointParms () const
					{ return (m_Point); }

				//! Get terrain parameters.
				const TERRAINPARMS& GetTerrainParms () const
					{ return (m_Terrain); }

				//! Get timestamp of previous classification to use.
				const DATETIME& GetTimeStampClassPrev () const
					{ return (m_TimeStampClassPrev); }

				//! Determine if references a single LAS file instead of LAS dataset.
				bool IsSingleLAS () const
					{ return (m_ObjItem.GetObjectType() == RVC::OBJTYPE_Shape); }

				//! Set to use specified object.
				ERRVALUE SetObject (
					const RVC::OBJITEM& objitem,		//!< Object to set
					bool keepsettings = false			//!< Retain current display settings
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				typedef LAYER_LIDAR::DISPPARM THISCLASS;

				static ERRVALUE SerializerCB_StyleFileName (SERIALIZER&, const SERIALIZERITEM *, void *, SERIALIZERITEM::ACTION);

				RVC::OBJITEM m_ObjItem;
				RVC::OBJECTNAME m_GeorefName;
				RVC::OBJITEM m_ObjItemStyle;
				POINTPARMS m_Point;
				TERRAINPARMS m_Terrain;
				DATETIME m_TimeStampClassPrev;

				void Copy (const DISPPARM&);
				void Free ();
				void GetFilePathDispParm (FILEPATH& filepath) const;

				//! Overrides from SERIALIZABLE.
				virtual ERRVALUE SerialRead (SERIALIZER& serializer);
				virtual ERRVALUE SerialWrite (SERIALIZER& serializer, const char *tagname = 0) const;

				//! Overrides from LAYER::DISPPARM.
				virtual const SERIALIZERITEM * SerialGetItemDef (SERIALIZER& serializer) const;
				virtual ERRVALUE v_SetObjItem (const RVC::OBJITEM& objitem, bool replace);
				virtual ERRVALUE v_Validate ();

				friend class LAYER_LIDAR;
				friend class LAYER_LIDAR::PRIV;
				#endif
			};	// End of LAYER_LIDAR::DISPPARM

		//! Register this layer type
		static void RegisterType ();

		//! Construct layer and insert into list.
		explicit LAYER_LIDAR (
			GRE::GROUP *group,									//!< Group to contain layer
			CREATEFLAGS createflags = CREATEFLAG_None,	//!< Creation flags
			GRE::LISTPOS listpos = GRE::LISTPOS_Last,		//!< Position in list
			LAYER *reflayer = 0									//!< Reference layer for insertion
			);

		//! Destructor.
		virtual ~LAYER_LIDAR ();

		//! Get reference to LAS dataset.
		LIDAR::LAS::DATASET& GetDatasetLAS () const;

		//! Retrieve layer-specific display parameters
		const DISPPARM& GetDispParm (
			) const { return (m_dispparm); }

		//! Get special color and size based on point record.
		void GetSpecialColorAndSize (
			const LIDAR::LAS::POINTRECORD& PointRec,
			GRE::VIEW *view,
			COLOR& color,
			UINT16& size
			) const;

		//! Set layer-specific display parameters.
		//! Validation will be performed on the specified parameters and
		//! adjustments made to the layer's copy if necessary.  Use of this method
		//! usually results in a ChangeBegin / ChangeEnd message pair, and will also
		//! generate a ChangeObject message if the spatial object used in the layer
		//! is changed.
		ERRVALUE SetDispParm (
			const DISPPARM& dispparm					//!< Parameters to set
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		DISPPARM m_dispparm;								// Layer-specific display parameters
		COMPINFO m_PointCompInfo;
	#ifdef USE_MGUI
		DISPPARM::DLG *m_pDlgControlPanel;
	#endif
		PRIV *m_pPriv;

		ERRVALUE DlgEditClassDescSet (MDLGPARENT dlgparent);
		const MISTRING& GetClassName (UINT16 ClassNumber) const;
		const STAT_UNIVARIATE& GetStatIntensity () const;
		ERRVALUE SetupLegend ();

		// Overrides from LAYER.
		virtual ERRVALUE v_AssignObject (const RVC::OBJITEM& objitem, MDLGPARENT dlgparent);
		virtual bool v_CanShowDetails () const;
		virtual double v_ComputeDftMaxVisibleScale () const;
		virtual GRE::LAYERSECTION* v_ConstructLayerSection (GRE::VIEW* view);
	#ifdef USE_MGUI
		virtual int v_ControlPanel (MDLGPARENT dlgparent, UINT32 flags = 0);
	#endif
		virtual LAYER* v_Copy (GRE::GROUP *group) const;
		virtual const DATATIPPARMS* v_DataTipGetParms (int component) const;
		virtual int v_DataTipGetString (ELEMTYPE ElemType, ELEMNUM ElemNum, MISTRING& string, SPATMOD::IMAGE::STAGE_ARRAY& images);
		virtual ERRVALUE v_DataTipSetParms (int component, const DATATIPPARMS& DataTipParms);
		virtual ERRVALUE v_Draw (LAYERDC& layerdc);
		virtual ERRVALUE v_DrawElementLow (LAYERDC& layerdc, ELEMTYPE ElemType, ELEMNUM ElemNum, const void *style);
		virtual ERRVALUE v_DrawElementsLow (LAYERDC& layerdc, ELEMTYPE ElemType, const UNSIGNEDSET *pElemSet, const void *style);
		virtual ERRVALUE v_EstimateRangeZ (const LAYERVIEW *pLayerView, DOUBLE_RANGE& range) const;
		virtual ELEMNUM v_FindClosestElement (const DPOINT2D& point, double searchdist, double *distret, int component, const UNSIGNEDSET *pSearchSet, GRE::FINDFLAGS findflags);
		virtual int v_GetComponent (ELEMTYPE ElemType) const;
		virtual MICONSTRUCT* v_GetComponentIcon (int component) const;
		virtual COMPINFO* v_GetComponentInfo (int component);
		virtual MISTRING v_GetComponentName (int component) const;
		virtual const void* v_GetElemDefaultStyle (ELEMTYPE ElemType) const;
		virtual UINT16 v_GetElemDrawSelectionMode (ELEMTYPE ElemType) const;
		virtual UINT16 v_GetElemDrawStyleMode (ELEMTYPE ElemType) const;
		virtual ERRVALUE v_GetElemExtents (ELEMTYPE ElemType, ELEMNUM ElemNum, DRECT3D& Extents);
		virtual ERRVALUE v_GetElemSetByExtent (ELEMTYPE ElemType, const DRECT2D& Extent, UNSIGNEDSET& ElemSet);
		virtual ELEMTYPE v_GetElemType (int component) const;
		virtual bool v_GetGeorefName (RVC::OBJECTNAME& name, int component = 0) const;
		virtual const GRE::LEGEND* v_GetLegend (ELEMTYPE ElemType);
		virtual INT64 v_GetNumElements (ELEMTYPE ElemType);
		virtual const RVC::OBJITEM& v_GetObjItem (int component) const;
		virtual ERRVALUE v_GetStyleInfo (ELEMTYPE ElemType, RVC::OBJITEM& ObjItem, RVC::OBJECTNAME& TableName);
		virtual bool v_HasDenseDetail () const;
		virtual bool v_HasDispParmSubObject () const;
		virtual bool v_HasLegend () const;
		virtual bool v_IsTableUsableForComponent (int component, const RVC::DBTABLE& table) const;
	#ifdef USE_MGUI
		virtual void v_MenuAddLayerItems (MGUI::MENU& menu);
		virtual ERRVALUE v_PromptObject (bool OpenCtrlPanel, MDLGPARENT dlgparent);
	#endif
		virtual ERRVALUE v_ReportRawElemInfo (MISTRING& infostr, ELEMTYPE ElemType, ELEMNUM ElemNum);
		virtual ERRVALUE v_SaveDispParmSubObject (RVC::DISPPARMUSAGE usage);
		virtual const char* v_SerialGetTagName () const;
		virtual ERRVALUE v_SerialRead (SERIALIZER& serializer);
		virtual void v_SetElemDrawSelectionMode (ELEMTYPE ElemType, UINT16 SelectMode);
		virtual void v_SetElemDrawStyleMode (ELEMTYPE ElemType, UINT16 StyleMode);
		virtual void v_SetLegendSampleFlags (MGD::STYLESAMPLEFLAGS SampleFlags);
		virtual void v_SetObjectDescriptor (int component, const RVC::DESCRIPTOR& descriptor);
		virtual ERRVALUE v_SetStyleObjItem (const RVC::OBJITEM& ObjItem);
		virtual int v_TestElementByRegion (ELEMTYPE ElemType, ELEMNUM ElemNum, const REGION2D& Region, TESTCRITERIA TestMode);
	#ifdef USE_MGUI
		virtual MGUI::CTRL_TREE::ITEMHND v_TreeViewAddUpdateLegend (GRE::CTRL_TREEVIEW& TreeView, MGUI::CTRL_TREE::ITEMHND hndParent, MGUI::CTRL_TREE::ITEMHND hndInsertAfter);
		virtual void v_TreeViewAddUpdateDetails (GRE::CTRL_TREEVIEW& TreeView, MGUI::CTRL_TREE::ITEMHND hndParent, MGUI::CTRL_TREE::ITEMHND hndInsertAfter);
	#endif
		virtual int v_UpdateExtents (bool forceupdate);

		// Unimplemented
		LAYER_LIDAR (const LAYER_LIDAR&);
		LAYER_LIDAR& operator= (const LAYER_LIDAR&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(LAYER_LIDAR::DISPPARM::RETURNSELECTION)
DEFINE_ENUM_OP_BITWISE(LAYER_LIDAR::DISPPARM::CLASSSELECTMODE)
#endif

}	// End namespace GRE

#endif	// INC_GRE_LRLIDAR_H

