/**
 * \file mgui/formlist.h
 * \brief MGUI::FORM_LIST class definitions
 *
 * \if NODOC
 * $Id: formlist.h_v 1.32 2012/11/26 16:05:04 mju Exp $
 *
 * $Log: formlist.h_v $
 * Revision 1.32  2012/11/26 16:05:04  mju
 * Add liststyle_NoInvertSelected.
 *
 * Revision 1.31  2012/11/21 16:47:02  mju
 * Comment out liststyle_useSelectButtons as not implemented.
 *
 * Revision 1.30  2012/08/28 16:31:59  mju
 * Add fieldstyle_IconTextRight/Right.
 *
 * Revision 1.29  2012/06/15 16:12:50  mju
 * Add liststyle_AllowMultiColumnDetails.
 *
 * Revision 1.28  2011/11/29 22:24:45  mju
 * Remove redundant hdr.
 *
 * Revision 1.27  2011/09/01 19:40:35  mju
 * No legacy macro.
 *
 * Revision 1.26  2011/08/05 15:32:51  mju
 * Add liststyle_ToggleSelect.
 *
 * Revision 1.25  2011/06/21 14:30:30  mju
 * Remove unneeded forward declarations.
 *
 * Revision 1.24  2011/04/20 19:01:57  mju
 * Add flashItem.
 *
 * Revision 1.23  2011/04/15 21:42:25  mju
 * Add liststyle_NoMultiColumn and lockLayout method.
 *
 * Revision 1.22  2011/04/13 20:00:52  mju
 * Add liststyle_NoFitToContent.
 *
 * Revision 1.21  2011/03/28 20:30:46  bcolombe
 * Added GetRowHeight
 *
 * Revision 1.20  2011/03/28 19:24:12  bcolombe
 * Add SetRect
 *
 * Revision 1.19  2011/03/07 18:05:22  mju
 * Add setRowHeight.
 *
 * Revision 1.18  2011/03/07 16:29:02  mju
 * Change customdraw::getRect to return reference.
 *
 * Revision 1.17  2011/01/19 23:47:38  bcolombe
 * Add delegate for editcancel
 *
 * Revision 1.16  2010/10/14 23:00:52  bcolombe
 * Added tooltip delegate that provides canvas point
 *
 * Revision 1.15  2010/10/07 14:42:52  mju
 * Add isFieldVisible.
 *
 * Revision 1.14  2010/09/30 13:08:47  mju
 * Add dlgShowHideFields.
 *
 * Revision 1.13  2010/02/17 17:14:24  mju
 * Add hasSortFields.
 *
 * Revision 1.12  2009/12/02 14:37:29  mju
 * Add ensurevisible_Beginning/End.
 *
 * Revision 1.11  2009/11/13 15:36:25  mju
 * Add mode to ensureVisible.
 *
 * Revision 1.10  2009/11/12 17:21:36  mju
 * Change to work with iconid directly instead of image list.
 * Add liststyle_UseIcons and setIconSize.
 *
 * Revision 1.9  2009/11/11 21:16:20  mju
 * Add liststyle_noSaveWidth/Height/Size.
 *
 * Revision 1.8  2009/10/28 19:53:16  mju
 * Add customdraw::hasSetBackground/TextColor.
 *
 * Revision 1.7  2009/10/15 19:59:55  mju
 * Add setDelegateOnCompare to use for all fields.
 *
 * Revision 1.4  2009/09/25 15:15:27  mju
 * Adjust how inikey/group specified and add saveLayout.
 * Remove minDisplayWidth from field creation.
 *
 * Revision 1.1  2009/09/18 17:49:35  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_MGUI_FORMLIST_H
#define	INC_MGUI_FORMLIST_H

#ifndef	INC_MGUI_CTRL_H
	#include <mgui/ctrl.h>
#endif

#ifndef	INC_MGUI_KEYCODE_H
	#include <mgui/keycode.h>
#endif

#ifndef	INC_MGUI_POINTERBUTTON_H
	#include <mgui/pointerbutton.h>
#endif

#ifndef	INC_MI32_MISTRING_H
	#include <mi32/mistring.h>
#endif

#ifndef	INC_MI32_COLOR_H
	#include <mi32/color.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
namespace MGD {
	class CONTEXT;
	}
struct LPOINT2D;
#endif


namespace MGUI {

//=====================================================================================================================
//! List control.
//! There are two general ways to work with lists, "standard" and "owner-data" (virtual).  In the
//! standard configuration the list control itself contains all of the data (strings, icons, states)
//! for each list entry.  In the "owner-data" configuration, the data is stored outside the list
//! and the OnGetDispInfo and OnOwnerDataCacheHint methods are overridden (or set when using the
//! template) and will be called by the list control when data is needed to be displayed in the list.
class FORM_LIST : public MGUI::FORM_COMPOSITE {
	#ifndef GENERATING_DOXYGEN_OUTPUT
	private:
		class PRIV;
	#endif // GENERATING_DOXYGEN_OUTPUT

	public:

		//! List style, specified in Create().
		enum LISTSTYLE {
			LISTSTYLE_NoSelect =						0x00000000,	//!< Don't allow selection by user (default)
			LISTSTYLE_SelectSingle =				0x00000001,	//!< Allow single selection
			LISTSTYLE_SelectMulti =					0x00000002,	//!< Allow multiple selection
			LISTSTYLE_WidthInPixels =				0x00000004,	//!< Width specified in pixels
			LISTSTYLE_HeightInPixels =				0x00000008,	//!< Height specified in pixels
			LISTSTYLE_UseCommonBackground =		0x00000010,	//!< Use common control background color instead of default white
			LISTSTYLE_NoFrame =						0x00000020,	//!< Don't include outer frame
			LISTSTYLE_NoHeading =					0x00000040,	//!< Don't show headings
			LISTSTYLE_NoScrollH =					0x00000080,	//!< Don't show horizontal scrollbar
			LISTSTYLE_NoScrollV =					0x00000100,	//!< Don't show vertical scrollbar
			LISTSTYLE_NoScrollBars =				LISTSTYLE_NoScrollH | LISTSTYLE_NoScrollV,	//!< Don't show any scrollbars
			LISTSTYLE_NoSort =						0x00000200,	//!< Don't allow sorting on any field
			LISTSTYLE_NoToolTips =					0x00000400,	//!< Don't show any tooltips
			LISTSTYLE_UseIcons =						0x00000800,	//!< Include space for icons when drawing (applies to field 0 only)
			LISTSTYLE_ShowItemSeparators =		0x00001000,	//!< Show separators between items
			LISTSTYLE_ShowFieldSeparators =		0x00002000,	//!< Show separators between fields in VIEWMODE_Details
			LISTSTYLE_ShowColumnSeparators =		0x00004000,	//!< Show separators between "columns" in multi-column mode
//			LISTSTYLE_UseSelectButtons =			0x00008000,	//!< Use buttons at left for selection
			LISTSTYLE_AllowReorderFields =		0x00010000,	//!< Allow field reorder by dragging column heading
			LISTSTYLE_AllowShowHideFields =		0x00020000,	//!< Allow user to show/hide fields
			LISTSTYLE_NoSaveWidth =					0x00040000,	//!< Don't read/write width from ini file.
			LISTSTYLE_NoSaveHeight =				0x00080000,	//!< Don't read/write height from ini file.
			LISTSTYLE_NoSaveSize =					LISTSTYLE_NoSaveWidth | LISTSTYLE_NoSaveHeight,	//!< Don't read/write size from ini file
			LISTSTYLE_NoMultiColumn =				0x00100000,	//!< Don't allow multi-column layout
			LISTSTYLE_ToggleSelect =				0x00200000,	//!< Toggle "selected" on normal left-click for LISTSTYLE_SelectMulti
			LISTSTYLE_AllowMultiColumnDetails =	0x00400000,	//!< Allow multi-column arrangement in "details" mode
			LISTSTYLE_NoFitToContent =				0x00800000,	//!< Don't allow "fit to content" (used for large "owner-data" lists)
			LISTSTYLE_NoInvertSelected =			0x01000000,	//!< Don't invert selected items
			};

		//! Field styles, specified in AddField().
		enum FIELDSTYLE {
			FIELDSTYLE_AlignLeft =				0x0000,	//!< Left-align text (default)
			FIELDSTYLE_AlignRight =				0x0001,	//!< Right-align text
			FIELDSTYLE_AlignCenter =			0x0002,	//!< Center text in column
			FIELDSTYLE_AlignMask =				0x0003,	//!< Mask for selecting alignment from style
			FIELDSTYLE_AllowEdit =				0x0004,	//!< Allow user edit of field value
			FIELDSTYLE_NoSort =					0x0008,	//!< Don't allow sorting on this field
			FIELDSTYLE_NoResize =				0x0010,	//!< Don't allow resize
			FIELDSTYLE_NoShowHide =				0x0020,	//!< Don't allow user to show/hide this field, only applies if LISTSTYLE_AllowShowHideFields set
			FIELDSTYLE_InitHidden =				0x0100,	//!< Field initially hidden by default
			FIELDSTYLE_InitSortDescending =	0x0200,	//!< Initially sort descending if use in sort
			FIELDSTYLE_IconTextRight =			0x0400,	//!< Place icon to immediate right of text
			FIELDSTYLE_IconRight =				0x0800,	//!< Place icon at right side of field
//			FIELDSTYLE_KeepWithPrevious =		0x0400,	//!< Keep field with previous one when reordering via user drag of column heading
//			FIELDSTYLE_NoSeparatorBefore =	0x0800,	//!< Don't show separator between this field and previous, implies FIELDSTYLE_KeepWithPrevious
			};

		enum VIEWMODE {
			VIEWMODE_List =			0,		//!< Only show first subitem for each item in multi-column arrangement
			VIEWMODE_Details =		1,		//!< Show in "grid" arrangement with each subitem in own column
			VIEWMODE_COUNT
			};

		//! Mode for starting edit of item field.
		enum EDITMODE {
			EDITMODE_InsertEnd,				//!< Don't select any text, insert at end
			EDITMODE_SelectAll,				//!< Select all text
			EDITMODE_Clear						//!< Clear text to start editing
			};

		//! Modes for ComputeFieldDisplayWidth().
		enum FIELDWIDTH {
			FIELDWIDTH_HeadingOnly =			0,
			FIELDWIDTH_HeadingAndData =		-1,
			FIELDWIDTH_DataOnly =				-2,
			FIELDWIDTH_AsAdded =					-3,
			};

		//! Enumeration of colors that can be set.
		enum COLORUSE {
			COLORUSE_HeadingBackground = 0,
			COLORUSE_HeadingText,
			COLORUSE_HeadingSeparator,
			COLORUSE_HeadingShadowDark,
			COLORUSE_HeadingShadowLight,
			COLORUSE_ItemBackgroundNormal,
			COLORUSE_ItemBackgroundSelected,
			COLORUSE_ItemTextNormal,
			COLORUSE_ItemTextSelected,
			COLORUSE_ItemSeparator,
			COLORUSE_FieldSeparator,
			COLORUSE_ColumnSeparator,
			COLORUSE_COUNT
			};

		enum ENSUREVISIBLE {
			ENSUREVISIBLE_ScrollMinimum =		0,		// Scroll list minimum amount to show item
			ENSUREVISIBLE_Beginning =			1,		// Scroll so item is at first visible position
			ENSUREVISIBLE_NearBeginning =		2,		// Scroll so item is near beginning of visible portion
			ENSUREVISIBLE_NearEnd =				3,		// Scroll so item is near end of visible portion
			ENSUREVISIBLE_End =					4,		// Scroll so item is at last visible position
			ENSUREVISIBLE_PartialOK =			9		// Don't scroll if item already partially visible
			};

		//! Structure passed to delegate for OnGetDispInfo.
		class DISPINFO {
			public:

				DISPINFO () :
					m_IconID(ICONID__NONE)
					{ }
					
				~DISPINFO() {}

				MISTRING m_text;				//!< String to display
				ICONID m_IconID;				//!< Icon ID
			};
		
		//! Passed to delegate for OnDrawItem().
		class CUSTOMDRAW {
			public:
				//! Get color for drawing background.
				const COLOR& GetColorBackground (
					) const { return (m_ColorBackground); }
				//! Get color for drawing text.
				const COLOR& GetColorText (
					) const { return (m_ColorText); }
				//! Get field number to draw.
				int GetFieldNum (
					) const { return (m_fieldnum); }
				MGD::CONTEXT* GetGC (
					) const { return (m_gc); }
				//! Get item number to draw.
				int GetItemNum (
					) const { return (m_itemnum); }
				//! Get rectangle on canvas for field, may not be entirely inside canvas.
				const LRECT2D& GetRect (
					) const { return (m_rect); }
				//! Determine if background color has been set.
				bool HasSetColorBackground (
					) const { return (m_ChangedColorBackground); }
				//! Determine if text color has been set.
				bool HasSetColorText (
					) const { return (m_ChangedColorText); }
				//! Set color for drawing background, use to override standard drawing.
				void SetColorBackground (
					const COLOR& color
					) { m_ColorBackground = color; m_ChangedColorBackground = true; }
				//! Set color for drawing text, use to override standard drawing.
				void SetColorText (
					const COLOR& color
					) { m_ColorText = color; m_ChangedColorText = true; }
				//! Set rectangle for field
				void SetRect (
					const LRECT2D& rect
					) { m_rect = rect; }
			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				friend class FORM_LIST::PRIV;

				int m_itemnum;							// Item number
				int m_fieldnum;						// Field number
				LRECT2D m_rect;						// Rectangle on canvas for entire field, may not be entirely inside canvas
				COLOR m_ColorBackground;
				COLOR m_ColorText;
				bool m_ChangedColorBackground;	// Delegate can set to true to indicate that background color was changed
				bool m_ChangedColorText;			// Delegate can set to true to indicate that text color was changed
				MGD::CONTEXT *m_gc;					// Drawing context if doing fully customized drawing

				CUSTOMDRAW (int itemnum, int fieldnum, MGD::CONTEXT *gc);
				~CUSTOMDRAW ();
				#endif
			};

		//! Delegate for pointer messages.
		//! Delegate should return 'true' if handled message, 'false' to continue with default handling.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum, const MGUI::POINT&, MGUI::POINTERBUTTON, MGUI::KEYSTATE)> DELEGATE_ONPOINTER;

		//! Delegate for Key presses.
		//! Should return true if keypress handled, false if not.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum, const MGUI::POINT&, MGUI::KEYSTATE, MGUI::KEYCODE)> DELEGATE_ONKEY;

		//! Delegate for item/field editing.
		//! Return true if editing allowed/accepted, false if not.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum, MISTRING&)> DELEGATE_ONEDIT;

		//! Delegate for item/field.  Used for OnEditCancel.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum)> DELEGATE_ONFIELD;

		//! Delegate for custom drawing of item/field.
		//! Return true if did all the drawing, false if not.
		typedef fastdelegate::FastDelegate<bool(CUSTOMDRAW& customdraw)> DELEGATE_ONCUSTOMDRAW;

		//! Delegate to get display information for item/field.
		typedef fastdelegate::FastDelegate<void(int itemnum, int fieldnum, DISPINFO&)> DELEGATE_ONGETDISPINFO;

		//! Delegate to compare for sorting.
		//! Return < 0 if item1 before item2, > 0 if item1 after item 2.
		typedef fastdelegate::FastDelegate<int(int itemnum1, int itemnum2, int fieldnum)> DELEGATE_ONCOMPARE;

		//! Delegate for control resize.
		typedef fastdelegate::FastDelegate<void(int width, int height)> DELEGATE_ONSIZE;

		//! Delegate to get tooltip for item/field.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum, MGUI::CTRL_TOOLTIP&)> DELEGATE_ONTOOLTIP;

		//! Delegate to get tooltip for item/field with canvas point.
		typedef fastdelegate::FastDelegate<bool(int itemnum, int fieldnum, MGUI::CTRL_TOOLTIP&, const MGUI::POINT&)> DELEGATE_ONTOOLTIPPOINT;

		//! Constructor.
		FORM_LIST ();

		//! Destructor.
		virtual ~FORM_LIST ();

		//! Add field to control.
		//! May be called before or after Create.  If used before Create() then may pass -1 for width
		//! in Create to automatically compute width to show all visible fieldnums in Report mode.
		//! @return fieldnum index.
		int AddField (
			const MISTRING& label,								//!< Label / column heading
			FIELDSTYLE style = FIELDSTYLE_AlignLeft,		//!< Field style
			int DisplayWidth = FIELDWIDTH_HeadingOnly,	//!< Display width in pixels, or FIELDWIDTH enum.
			const char *IniKey = 0								//!< Key for saving settings, NULL if not saveable (ie, can't be changed)
			);

		//! Append item at end of list.
		//! @return Item number of inserted item if successful, < 0 if error.
		int AppendItem (
			const MISTRING& string						//!< String to assign to first field
			);

		//! Compute and set width of field based on heading and/or data.
		//! @see SetFieldDisplayWidth().
		void ComputeFieldDisplayWidth (
			int fieldnum,									//!< Field number
			FIELDWIDTH mode								//!< Mode
			);

		//! Compute point relative to item/field rectangle.
		//! @return true if canvas point inside field rectangle, false if not.
		bool ComputeRelativePoint (
			int itemnum,									//!< Item number
			int fieldnum,									//!< Field number
			const MGUI::POINT& ptCanvas,				//!< Point on canvas
			MGUI::POINT& ptRelative						//!< Point relative to item/field rectangle returned
			) const;

		//! Create the control.
		//! It is recommended that fields be added before calling Create().
		//! Note that width/height values do not include heading or scrollbars, if any.
		//! If a key is provided it will be used to read the size and field order (if fields can be reordered and are already added).
		void Create (
			MGUI::LAYOUT_PANE_BASE& ParentPane,		//!< Parent pane
			int height,										//!< Height of list area in rows or pixels if LISTSTYLE_HeightInPixels
			int width,										//!< Width of list area in 'em' characters or pixels if LISTSTYLE_WidthInPixels, -1 to compute if fields already added
			LISTSTYLE style = LISTSTYLE_NoSelect,	//!< List style
			VIEWMODE ViewMode = VIEWMODE_Details,	//!< Initial viewing mode
			MGUI::LAYOUT_SIZEALIGN sizealign = MGUI::LAYOUT_SIZEALIGN_Expand
			);

		//! Delete all fields.
		void DeleteAllFields (
			);

		//! Delete all items.
		void DeleteAllItems (
			) { SetItemCount(0); }

		//! Delete specified field.
		//! Note that first field cannot be deleted.
		void DeleteField (
			int fieldnum
			);

		//! Delete specified item.
		void DeleteItem (
			int itemnum
			);

		//! Delete specified item given assigned data value.
		void DeleteItemByID (
			ID id
			) { DeleteItem(FindItemByID(id)); }

		//! Deselect all items.
		void DeselectAllItems (
			bool notify = true
			);

		//! Open dialog to allow user to choose fields to show/hide.
		//! @return true if changed, false if not.
		bool DlgShowHideFields (
			MDLGPARENT dlgparent
			);

		//! Begin editing specified item field.
		//! @return true if successful, false if not.
		bool EditItem (
			int itemnum,
			int fieldnum,
			EDITMODE editmode = EDITMODE_InsertEnd				//!< Mode to start editing item
			);

		//! Scroll list to ensures that an item is visible.
		void EnsureVisible (
			int itemnum,
			int fieldnum = -1,
			ENSUREVISIBLE mode = ENSUREVISIBLE_ScrollMinimum
			);

		//! Find item by matching specified data value as ID.
		//! @return Item number or -1 if none found.
		int FindItemByID (
			ID id
			) const;

		//! Find item by matching specified data value as pointer.
		//! @return Item number or -1 if none found.
		int FindItemByID (
			const void *id
			) const { return (FindItemByID(reinterpret_cast<ID>(id))); }

		//! Flash item in list.
		void FlashItem (
			int itemnum,
			int numtimes = 1
			) const;

		//! Get color used in interface.
		const COLOR& GetColor (
			COLORUSE use
			) const;

		//! Get number of fields in the list.
		int GetFieldCount (
			) const;

		//! Get field display order in "details" mode.
		void GetFieldDisplayOrderDetails (
			SIMPLE_ARRAY<int>& DisplayOrder
			) const;

		//! Get the width to display field in pixels.
		int GetFieldDisplayWidth (
			int fieldnum
			) const;

		//! Get the label for a field.
		//! @return true if successful, false otherwise
		MISTRING GetFieldLabel (
			int fieldnum
			) const;

		//! Get first visible item number.
		int GetFirstVisibleItem (
			) const;

		//! Get number of items in list.
		int GetItemCount (
			) const;

		//! Get data value associated with item.
		MGUI::ID GetItemID (
			int itemnum
			) const;

		//! Get text assigned to item/field.
		MISTRING GetItemText (
			int itemnum,
			int fieldnum
			) const;

		//! Get parent for popup menu
		MGUI::CTRL& GetMenuParent (
			) const;

		//! Get next selected item in list.
		//! @return Index of next item or -1 if no more items.
		int GetNextSelectedItem (
			int startitem = -1							//!< Starting item for search, starting item is excluded
			) const;

		//! Get number of visible rows.
		int GetNumVisibleRows (
			) const;

		//! Get number of selected items.
		int GetSelectedCount (
			) const;

		//! Get size.
		const LPOINT2D& GetSize (
			) const;

		//! Get height of row
		int GetRowHeight (
			) const;

		//! Determine if list has been sorted.
		//! Typically used after adding items to determine if should sort.
		bool HasSorted (
			) const;

		//! Determine if has specified fields to sort on.
		bool HasSortFields (
			) const;

		//! Insert item at specified position.
		//! @return Item number of inserted item if successful, < 0 if error.
		int InsertItem (
			int pos,											//!< Item position in list
			const MISTRING& string						//!< String to assign to first field
			);

		//! Determine if field is sorted ascending or not.
		//! Can be used for specialized sorting behaviors.
		bool IsFieldSortAscending (
			int fieldnum
			) const;

		//! Determine if field is visible in specified or current viewing mode.
		bool IsFieldVisible (
			int fieldnum,
			VIEWMODE ViewMode = VIEWMODE_COUNT		//! Viewing mode, VIEWMODE_COUNT to use current
			) const;

		//! Determine if specified item is currently "selected".
		//! @return 'true' if selected, 'false' if not.
		bool IsItemSelected (
			int itemnum
			) const;

		//! Lock layout to another list.
		//! Any number of lists can be locked together.
		//! Changes to layout for one (field reorder, sizing, visibility) will affect all other locked lists.
		//! Sizing a field to content will affect all lists.
		void LockLayout (
			FORM_LIST& ListLockTo
			);

		//! Redraw all items.
		void Redraw (
			);

		//! Disable redraw, incrementing disable count.
		void RedrawDisable (
			);

		//! Enable redraw, decrementing disable count.
		//! Will redraw control if was disabled and is no longer.
		void RedrawEnable (
			);

		//! Redraw specified item.
		void RedrawItem (
			int itemnum
			);

		//! Redraw range of items.
		void RedrawItems (
			int firstitemnum,
			int lastitemnum
			);

		//! Save layout to Ini file based on keys previously provided in AddField and Create.
		void SaveLayout (
			) const;

		//! Set color to use in interface.
		//! Note that non-background colors can include a transparency setting that will be used to blend with
		//! any previously set or default background color.  If the background color is changed after any blending
		//! is done the previously blended color will be retained.
		void SetColor (
			COLORUSE use,									//!< Color use
			const COLOR& color							//!< Color to set
			);

		//! Set size of icons to use.
		//! If used, must be called before Create().
		void SetIconSize (
			int IconSize									//! Icon size
			);

		//! Set field display order for "details" mode.
		//! Each field must be listed in the array exactly once and all fields must be listed.
		void SetFieldDisplayOrderDetails (
			const SIMPLE_ARRAY<int>& FieldOrder
			);

		//! Set width of a field for display.
		//! @see ComputeFieldDisplayWidth().
		void SetFieldDisplayWidth (
			int fieldnum,
			int width										//!< Width in pixels
			);

		//! Set field label.
		void SetFieldLabel (
			int fieldnum,
			const MISTRING& label
			);

		//! Set whether field column is visible in 'details' mode.
		//! Note that field 0 cannot be hidden
		void SetFieldVisible (
			int fieldnum,
			bool visible
			);

		//! Set delegate to call when some aspect of layout (column order, width, visibility etc) is changed.
		void SetDelegateOnChangeLayout (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Set delegate to call when order of items in list is changed.
		void SetDelegateOnChangeOrder (
			DELEGATE_VOID_NOPARMS delegate
			);

		//! Set delegate to call when selected items changed.
		//! If a specific item was used to initiate the change, either as part of a single or multiple
		//! selection action it will be passed to the delegate, otherwise -1 will be passed.  The item
		//! passed may or may not be "selected" depending on the user action performed.
		void SetDelegateOnChangeSelection (
			DELEGATE_VOID_INT32 delegate
			);

		//! Set delegate for comparing field from two items to use in sorting.
		//! Will only be used if no specific delegate set for field being compared.
		void SetDelegateOnCompare (
			DELEGATE_ONCOMPARE delegate
			);

		//! Set delegate for customized drawing.
		void SetDelegateOnDrawItem (
			DELEGATE_ONCUSTOMDRAW delegate
			);

		//! Set delegate to call before editing item/field.
		void SetDelegateOnEditBegin (
			DELEGATE_ONEDIT delegate
			);

		//! Set delegate to call for editing item/field cancel.
		void SetDelegateOnEditCancel (
			DELEGATE_ONFIELD delegate
			);

		//! Set delegate to call after editing item/field.
		void SetDelegateOnEditEnd (
			DELEGATE_ONEDIT delegate
			);

		//! Set delegate for comparing field from two items to use in sorting.
		void SetDelegateOnFieldCompare (
			int fieldnum,								//!< Field number to set delegate for
			DELEGATE_ONCOMPARE delegate
			);

		//! Set delegate to get display information for item/field.
		void SetDelegateOnGetDispInfo (
			DELEGATE_ONGETDISPINFO delegate
			);

		//! Set the delegate to call on a key press
		void SetDelegateOnKeyPress (
			DELEGATE_ONKEY delegate
			);

		//! Set delegate for double-click of left pointer button on item/field.
		void SetDelegateOnLeftButtonDoubleClick (
			DELEGATE_ONPOINTER delegate
			);

		//! Set the delegate to call on pointer button down
		//! Delegate should return 'true' to accept the default behavior.
		void SetDelegateOnPointerButtonDown (
			DELEGATE_ONPOINTER delegate
			);

		//! Set the delegate to call on pointer button up
		//! Delegate should return 'true' to accept the default behavior.
		void SetDelegateOnPointerButtonUp (
			DELEGATE_ONPOINTER delegate
			);

		//! Set delegate for control resize.
		void SetDelegateOnSize (
			DELEGATE_ONSIZE delegate
			);

		//! Set delegate to get ToolTip.
		void SetDelegateOnToolTip (
			DELEGATE_ONTOOLTIP delegate
			);

		//! Set delegate to get ToolTip with canvas point.
		void SetDelegateOnToolTipPoint (
			DELEGATE_ONTOOLTIPPOINT delegate
			);

		//! Set INI key and group for read/write settings.
		//! If used, must be done before any fields are added, and before Create.
		void SetIniKey (
			const char *IniKey,
			const char *IniGroup = 0
			);

		//! Set number of items in list.
		ERRVALUE SetItemCount (
			int NumItems									//!< Number of items
			);

		//! Set application-specific data to associate with list item.
		ERRVALUE SetItemID (
			int itemnum,									//!< Item number
			ID id												//!< Value to associate with item
			);

		//! Set icon ID to use for a givem item.
		ERRVALUE SetItemIcon (
			int itemnum,									//!< Item number
			ICONID iconid									//!< Icon ID
			);

		//! Set whether item is 'selected or not.
		void SetItemSelected (
			int itemnum,									//!< Item number
			bool selected = true,
			bool notify = true
			);

		//! Set item/field text.
		void SetItemText (
			int itemnum,									//!< Item number
			int fieldnum,									//!< Sub item, 0 for item
			const MISTRING& string						//!< Item or fieldnum string
			);

		//! Set row height.
		void SetRowHeight (
			int RowHeight,									//!< Row height in pixels
			VIEWMODE ViewMode = VIEWMODE_List
			);

		//! Set viewing mode.
		void SetViewMode (
			VIEWMODE ViewMode
			);

		//! Sort the data in the list using same order as previous.
		//! This is useful if items have been added to the list.
		//! If no previous sort has been performed the items are sorted using the first sortable field(s).
		void Sort (
			);

		//! Sort the data in the list on a given field.
		//! This will move the specified field to the front of the list of fields to sort on.
		void Sort (
			int fieldnum,									//!< Field to sort on
			bool bAscending = true
			);

		//! Sort data in the list using specified fields.
		void Sort (
			const SIMPLE_ARRAY<int>& FieldNums,				//!< Fields to sort on
			const SIMPLE_ARRAY<bool>& FieldsAscending		//!< Whether field should be sorted in ascending or descending order
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		friend class PRIV;
		PRIV *m_pPriv;

		// UNIMPLEMENTED / UNSUPPORTED
		FORM_LIST (const FORM_LIST&);
		FORM_LIST& operator= (const FORM_LIST&);

		#endif // GENERATING_DOXYGEN_OUTPUT

	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MGUI::FORM_LIST::LISTSTYLE)
DEFINE_ENUM_OP_BITWISE(MGUI::FORM_LIST::FIELDSTYLE)
#endif

//=====================================================================================================================

}	// End of MGUI namespace

#endif	// INC_MGUI_FORMLIST_H
