/**
 * \file mgui/toolinsertion.h
 * \brief MGUI::TOOL_INSERTION class definitions
 *
 * \if NODOC
 * $Id: toolinsertion.h_v 1.7 2013/02/12 22:13:53 bcolombe Exp $
 *
 * $Log: toolinsertion.h_v $
 * Revision 1.7  2013/02/12 22:13:53  bcolombe
 * Add GetTransformation that is computed from layer coordinates
 *
 * Revision 1.6  2011/12/09 20:23:08  mju
 * No context in v_draw.
 *
 * Revision 1.5  2008/11/14 20:39:57  scowan
 * Added setAngle method.
 *
 * Revision 1.4  2008/11/13 21:40:01  scowan
 * Added getInsertionPoint.
 *
 * Revision 1.3  2008/04/17 14:55:21  scowan
 * Manual entry panel.
 *
 * Revision 1.2  2008/04/16 23:01:22  scowan
 * Tool mostly working.
 *
 * Revision 1.1  2008/03/27 22:00:44  scowan
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MGUI_TOOLINSERTION_H
#define	INC_MGUI_TOOLINSERTION_H

#ifndef	INC_MGUI_TOOLBASE_H
#include <mgui/toolbase.h>
#endif

class TRANS2D_AFFINE;

namespace MGUI {

//! Insertion tool class.
class TOOL_INSERTION : public TOOL_BASE {
	public:

		//! Constructor.
		explicit TOOL_INSERTION (
			TOOLCANVAS *canvas
			);

		//! Destructor.
		virtual ~TOOL_INSERTION ();

		//! Get rotation angle.
		//! @return Rotation angle in RADIANS.
		double GetAngle (
			) const { return (m_angle); }

		//! Get extents from the current transformation to the destination based on the current tool settings.
		const DRECT2D& GetExtents (
			) const { return (m_extents); }

		//! Get the insertion point for the transformation
		//! The Insertion Point, combined with the source Base Point, Scale factors and Rotation, determine the affine transformation.
		const DPOINT2D& GetInsertionPoint (
			) const { return (m_InsertPoint); }
			
		//! Get current scale factors
		//! @return Current X and Y scale factors
		const DPOINT2D& GetScale (
			) const { return (m_scale); }
			
		//! Get the affine transformation from source to destination based on the current tool settings.
		//! This transformation is based on view coordinates
		//! @return The current affine transformation from source to destination coordinates.
		const TRANS2D_AFFINE& GetTransformation (
			) const { return (m_SrcToDest); }

		//! Get the affine transformation from source to destination based on the current tool settings.
		//! This transformation is based on layer coordinates
		void GetTransformation (
			const TRANS2D_MAPGEN& LayerToView,	//!< Transformation used to convert the box's view coordinates to layer coordinates
			TRANS2D_AFFINE& SrcToDest
			) const;

		//! Set rotation angle.
		void SetAngle (
			double angle
			);

		//! Set whether rotation angle indicator is visible.
		//! The rotation angle is indicated by a line from the center to the boundary.
		//! Only valid if the source base point is in the center of the box.
		void SetAngleVisible (
			bool anglevisible = true
			);

		//! Set whether center marker is visible.
		void SetCenterVisible (
			bool centervisible = true
			);

		//! Set the insertion point for the transformation
		//! The Insertion Point, combined with the source Base Point, Scale factors and Rotation, determine the affine transformation.
		void SetInsertionPoint (
			const DPOINT2D& InsertionPoint
			);
			
		//! Set current X and Y scale factors
		//! The Scale factors, combined with the source Base Point, Insertion Point and Rotation, determine the affine transformation.
		void SetScale (
			const DPOINT2D& scale
			);
			
		//! Set the 'source' position.
		void SetSourcePosition (
			const DRECT2D& SrcExtents,					//!< Extents of the source in tool coordinates
			const DPOINT2D* SrcBasePoint = 0,		//!< If not specified, the base point will be the the center point of the source extents
			bool ResetToIdentity = true				//!< If 'true', the transformation parameters will be reset to an identity transformation.
			);

		//! Set length of "side markers".
		//! By default, no side markers are shown.
		//! This is useful if box indicates a target if it can become small and hard to see.
		void SetSideMarkerLength (
			int lengthoutside								//!< Length of marker outside box in pixels
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		// Enumeration constants
		enum ZONE {
			ZONE_Outside = 0,
			ZONE_Inside,
			ZONE_Corner0,
			ZONE_Corner1,
			ZONE_Corner2,
			ZONE_Corner3,
			ZONE_Side01,
			ZONE_Side12,
			ZONE_Side23,
			ZONE_Side30,
			ZONE_Rotate,
			ZONE_SetCenter,
			ZONE_NoFunction
			};
		// Tool settings
		bool m_centervisible;					// Visibility of center indicator
		bool m_anglevisible;						// Visibility of rotation angle indicator
		int m_sidemarkerlength;					// Length in pixels of side markers out of box
		// Tool position values
		DRECT2D m_extents;						// Position (extents if rotated)
		DPOINT2D m_destCorners[4];				// Corner coordinates
		DPOINT2D m_scale;							// Current scale
		DPOINT2D m_BasePoint;					// Current source base point
		DPOINT2D m_InsertPoint;					// Current destination base point
		double m_angle;							// Rotation angle
		TRANS2D_AFFINE m_SrcToDest;			// Source to destination transformation
		// Values used during interaction
		ZONE m_PointerZone;						// Current cursor zone
		DPOINT2D m_prevpt;						// Previous point during button-down
		UINT8 m_AdjustMode;						// Adjustment mode, 2=rotate, 1=everything else

		//! Manual Entry form forward declaration
		class MANUAL_INSERTION;
		friend class MANUAL_INSERTION;

		// Methods
		void AdjustTool (const DPOINT2D&, bool firsttime=false);
		DPOINT2D ComputeCenter () const;
		void ComputePointerZone (const DPOINT2D&);
		void Recompute ();
		void ResizeTool (const DPOINT2D&);
		void RotateTool (const DPOINT2D&);
		void StartTool (const DPOINT2D&);
		void UpdateCursor (const DPOINT2D&) const;

		// TOOL_BASE overrides.
		virtual MANUAL_BASE* v_CreateManualBase ();
		virtual void v_DoDraw ();
		virtual void v_DoMove (double dx, double dy, ADJUSTFLAGS flags);
		virtual bool v_OnPointerButtonDown (MGUI::POINT point, POINTERBUTTON button, KEYSTATE keystate);
		virtual void v_OnPointerMove (MGUI::POINT point, POINTERBUTTON button, KEYSTATE keystate);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};


}	// End namespace MGUI


#endif	// INC_MGUI_TOOLRECTANGLE_H
