/**
 * \file mi32/btreefun.h
 * \brief Definitions for BTree functions
 *
 * \if NODOC
 * $Log: btreefun.h_v $
 * Revision 1.34  2012/01/27 22:04:31  mju
 * Move balancedtree template to own hdr.
 *
 * Revision 1.33  2012/01/27 21:51:33  mju
 * Use const data in testfunc.
 *
 * Revision 1.32  2011/12/14 17:43:44  mju
 * Remove unused methods.
 *
 * Revision 1.31  2011/12/13 15:20:24  mju
 * Remove unused fn.
 *
 * Revision 1.30  2006/09/18 14:35:27  scowan
 * Added necessary include file.
 *
 * Revision 1.29  2005/12/29 21:35:22  scowan
 * Removed the reference.
 *
 * Revision 1.28  2005/12/29 20:31:58  scowan
 * Added code to allocate a test and delete when done.
 *
 * Revision 1.27  2005/08/29 19:14:07  scowan
 * Fixed deprecated item.
 *
 * Revision 1.26  2003/12/18 13:59:55  mju
 * Remove deprecated items.
 *
 * Revision 1.25  2003/09/15 13:49:56  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.24  2003/09/08 20:33:39  dwilliss
 * Added doxygen tags for start/end of function groups
 *
 * Revision 1.23  2003/05/12 14:05:21  mju
 * Fix deprecated macro.
 *
 * Revision 1.22  2003/05/12 13:48:05  mju
 * Deprecate obsolete Init/Stop fns.
 *
 * Revision 1.21  2003/04/09 15:10:17  scowan
 * Fixed g++ 3.0 unix compiler error for mac osx.
 *
 * Revision 1.20  2003/03/20 15:48:17  scowan
 * Added is valid method.
 *
 * Revision 1.19  2003/01/27 17:16:22  scowan
 * Fixed more typo's.
 *
 * Revision 1.18  2003/01/27 16:44:56  scowan
 * Fixed wrapper class issues.
 *
 * Revision 1.17  2003/01/22 16:54:16  scowan
 * Fixed compile errors with wrapper class.
 *
 * Revision 1.16  2003/01/22 15:59:13  scowan
 * Added inline template wrapper, do not know if it works.
 *
 * Revision 1.15  2001/01/23 16:48:13  dwilliss
 * Had forgotten to export new function from the DLL
 * \endif
**/

#ifndef	INC_MI32_BTREEFUN_H
#define	INC_MI32_BTREEFUN_H

#ifndef	INC_MI32_STDDEFNS_H
#include <mi32/stddefns.h>
#endif

#ifdef GEOMDLL
	#define GEOMLIBEXPORT MI_DLLEXPORT
#else
	#define GEOMLIBEXPORT MI_DLLIMPORT
#endif

//----------------------------------------------------------------------------
//!		Function prototypes
//----------------------------------------------------------------------------

//!\addtogroup btree Btree Functions
//!@{

//! Tree comparison function used in searching/adding.
//! @return >, ==, < 0.
//! Must NOT modify the data.
typedef int (*BTreeTestFunc) (
	const void *,
	const void *,
	void *
	);

//! Callback used to return data. 
typedef int (*BTreeDumpFunc)(
	void *data, 								//!< Data stored in the tree
	INT32 *tagnum, 							//!< Pointer to tag number
	void *userdata								//!< Pointer to sent data
	);

#if defined(__cplusplus)
extern "C" {
#endif

//! Allocate a "balanced tree".
//! @return Handle to tree instance.										
GEOMLIBEXPORT void *BTreeAlloc (
	int numitems, 								//!< Number of items to store per node 
	int datasize, 								//!< Size of data to tree in bytes
	BTreeTestFunc testfunc,					//!< Comparison function used in searching/adding, MUST NOT MODIFY DATA
	void *userdata								//!<	Data passed to comparison function
	);

//! Clear "balanced tree".
GEOMLIBEXPORT int BTreeClear (
	void *tree	 								//!< Handle to tree instance
	);

//! Dump a "balanced tree" in "traversal" order.
//! NOTE: This function transverses the tree and dumps the data in the
//!	order that the comparison function determined.
GEOMLIBEXPORT int BTreeDumpT (
	void *tree, 								//!< Handle to tree instance
	BTreeDumpFunc, 							//!< Callback used to return data
	void *userdata								//!< User data passed to Dump()
	);

//! Search "balanced tree" for a specified key, add if no match found.
//! @return tag number assigned to the data if (>=0), else error
GEOMLIBEXPORT INT32 BTreeFindKey (
	void *tree, 								//!< Handle to tree instance
	void *data, 								//!< Data to find and add if necessary
	INT32 tagnum, 								//!< Number to tag data with
	int *add										//!< Set if data was added to tree
	);

//! Free a "balanced tree".
GEOMLIBEXPORT int BTreeFree (
	void *tree									//!< Handle to tree instance
	);

//! Get the userdata associated with 'tree'
GEOMLIBEXPORT void* BTreeGetUserData (
	void *tree	 								//!< Handle to tree instance
	);

//! Search "balanced tree" for a specified key.
//!
//! @return -1 = Not found, < -1 Error, else the tag number associated with the data.
//!	See BTreeFindKey(). 
GEOMLIBEXPORT INT32 BTreeSearchKey (
	void *tree, 								//!< Handle to tree instance
	void *data									//!< Data to find if exists
	);

#if defined(__cplusplus)
} // extern "C"
#endif

//!@}

#endif
