/**
 * \file mi32/kmlwriter.h
 * \brief MIKML::WRITER definitions
 *
 * \if NODOC
 * $Id: kmlwriter.h_v 1.9 2013/01/04 15:44:43 mju Exp $
 *
 * $Log: kmlwriter.h_v $
 * Revision 1.9  2013/01/04 15:44:43  mju
 * Add name/visibility to ...OverlayBegin.
 * Add vec2 parms to screenOverlayBegin.
 *
 * Revision 1.8  2013/01/04 14:33:09  mju
 * Add parms to document/folderBegin to create style/liststyle sections.
 *
 * Revision 1.7  2013/01/04 14:11:14  mju
 * Fix parm name.
 *
 * Revision 1.6  2013/01/03 22:40:59  mju
 * Add styleAdd and styleAddShared for xxxSTYLE support and shared styles.
 * Add ability to turn off use of newlines.
 *
 * Revision 1.5  2013/01/03 14:01:38  mju
 * Add styleBegin without id.
 *
 * Revision 1.4  2012/12/21 15:42:53  mju
 * Add schema and extendedData support.
 *
 * Revision 1.3  2012/12/17 16:21:10  mju
 * Add drawOrder to groundOverlayBegin.
 *
 * Revision 1.2  2012/12/17 16:15:58  mju
 * Move to MIKML namespace and add more methods.
 *
 * Revision 1.1  2012/12/14 21:28:08  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_KMLWRITER_H
#define	INC_MI32_KMLWRITER_H

#ifndef	INC_MI32_KMLBASE_H
#include <mi32/kmlbase.h>
#endif

#ifndef	INC_MI32_FILETEXT_H
#include <mi32/filetext.h>
#endif

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifndef	INC_MI32_MISTRING_H
#include <mi32/mistring.h>
#endif

#ifndef	INC_MI32_POINT_H
#include <mi32/point.h>
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
struct COLOR;
class POLYLINE;
class REGION2D;
class SUBREGION2D;
struct POINTSTYLE;
struct LINESTYLE;
struct POLYSTYLE;
struct TEXTSTYLE;
#endif


namespace MIKML {
//=====================================================================================================================

//! Sequential KML file writer.
//! Provides methods for writing most KML elements.
//!
//! Methods are provided for simple managing of shared styles.  To use shared styles, after DocumentBegin use
//! the StyleAddShared methods to build the set of styles.  Management of duplicate styles is automatic so there
//! is no need to check for duplicates first.  Then, as each Placemark is added, use the appropriate StyleAdd method
//! (without specifying the ID) and the previously added shared style will be located and a StyleUrl item will be
//! added for the Placemark.  Because the same method may be used when adding styles to a Placemark whether or not
//! shared styles are used, support can easily be added after initial development is done.
class WRITER {
	public:

		//! How to use shared styles in StyleAdd.
		enum SHAREDSTYLEUSE {
			SHAREDSTYLEUSE_None = 0,				//!< Don't use shared style
			SHAREDSTYLEUSE_IfAvailable,			//!< Use shared style if available, Feature-specific style if not
			SHAREDSTYLEUSE_Expected					//!< Expected to have shared style, return error if not available
			};

		//! Constructor.
		WRITER ();

		//! Destructor.
		~WRITER ();

		//! Add boolean value (stored as 0 or 1).
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddBoolean (
			const char *name,							//!< Element tag name
			bool value									//!< Value false=0, true=1
			);

		//! Add color (stored as AABBGGRR).
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddColor (
			const char *name,							//!< Element tag name
			const COLOR& color						//!< Color
			);

		//! Add floating-point value.
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddFloat (
			const char *name,							//!< Element tag name
			double value,								//!< Value
			int precision								//!< Maximum precision to write (will have trailing 0's stripped)
			);

		//! Add integer value.
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddInteger (
			const char *name,							//!< Element tag name
			INT32 value									//!< Value
			);

		//! Add string value.
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddString (
			const char *name,							//!< Element tag name
			const char *value							//!< Value string
			);

		//! Add string value.
		//! Does not perform any schema validity check on element tag.
		ERRVALUE AddString (
			const char *name,							//!< Element tag name
			const MISTRING& value					//!< Value string
			);

		//! Add "altitudeMode" item.
		ERRVALUE AltitudeModeAdd (
			ALTITUDEMODE altitudeMode				//!< Altitude mode
			);

		//! Add "BalloonStyle" element.
		ERRVALUE BalloonStyleAdd (
			const MISTRING& text,					//!< Balloon text template, usually should be wrapped in CDATA
			const COLOR *pTextColor = 0,			//!< Text color override
			const COLOR *pBgColor = 0,				//!< Background color override
			bool displayModeHide = false			//!< Whether to set displayMode to "hide"
			);

		//! Add "Camera" element.
		ERRVALUE CameraAdd (
			const DPOINT3D& ptCamera,				//!< Longitude, Latitude, Altitude of camera
			double heading,							//!< Heading degrees
			double tilt,								//!< Tilt degrees
			double roll = 0,							//!< Roll degrees
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround	//!< Altitude mode
			);

		//! Close KML file with validation.
		ERRVALUE Close ();

		//! Add "Data" (untyped name/value pair) element.
		ERRVALUE DataAdd (
			const MISTRING& name,					//!< Name attribute
			const MISTRING& value,					//!< Value
			const MISTRING *pDisplayName = 0		//!< Optional displayName string
			);

		//! Add "Data" (untyped name/value pair) element.
		ERRVALUE DataAdd (
			const char *name,							//!< Name attribute
			const MISTRING& value,					//!< Value
			const char *DisplayName = 0			//!< Optional displayName string
			);

		//! Add "description" element.
		ERRVALUE DescriptionAdd (
			const MISTRING& description			//!< Description string
			);

		//! Begin "Document" section.
		//! A "Style" and "ListStyle" section will be automatically added if non-default values are specified for listItemType or pListBgColor.
		ERRVALUE DocumentBegin (
			bool open = false,						//!< Whether to initially show as open (expanded) in list
			LISTITEMTYPE listItemType = LISTITEMTYPE_Check,		//!< How to show in list view
			const COLOR *pListBgColor = 0			//!< Background color in list view
			) { return (ContainerBegin(ITEMTYPE_Document,open,listItemType,pListBgColor)); }

		//! End "Document" section.
		ERRVALUE DocumentEnd ()
			{ return (AggregateEnd(ITEMTYPE_Document)); }

		//! Begin "ExtendedData" section.
		ERRVALUE ExtendedDataBegin (const MISTRING& id)
			{ return (AggregateBegin(ITEMTYPE_ExtendedData,"id",id)); }

		//! End "ExtendedData" section.
		ERRVALUE ExtendedDataEnd ()
			{ return (AggregateEnd(ITEMTYPE_ExtendedData)); }

		//! Begin "Folder" section.
		//! A "Style" and "ListStyle" section will be automatically added if non-default values are specified for listItemType or pListBgColor.
		ERRVALUE FolderBegin (
			bool open = false,						//!< Whether to initially show as open (expanded) in list
			LISTITEMTYPE listItemType = LISTITEMTYPE_Check,		//!< How to show in list view
			const COLOR *pListBgColor = 0			//!< Background color in list view
			) { return (ContainerBegin(ITEMTYPE_Document,open,listItemType,pListBgColor)); }

		//! End "Folder" section.
		ERRVALUE FolderEnd ()
			{ return (AggregateEnd(ITEMTYPE_Folder)); }

		//! Begin "GroundOverlay" section.
		ERRVALUE GroundOverlayBegin (
			const MISTRING& name,					//!< Name string
			const MISTRING& href,					//!< Location of linked item
			const DRECT2D& LatLonBox,				//!< Extent in degrees
			bool visibility = true,					//!< Whether initially visible
			int drawOrder = 0,						//!< Drawing Order
			double altitude = 0,						//!< Altitude
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround
			);

		//! End "GroundOverlay" section.
		ERRVALUE GroundOverlayEnd ()
			{ return (AggregateEnd(ITEMTYPE_GroundOverlay)); }

		//! Add "IconStyle" section.
		ERRVALUE IconStyleAdd (
			const COLOR& color,						//!< Icon color
			const MISTRING& href,					//!< Icon href
			double scale = 1,							//!< Resizing scale
			double heading = 0,						//!< Heading in degrees
			double hotSpotX = 0.5,					//!< Hotspot X
			double hotSpotY = 0.5,					//!< Hotspot Y
			UNITS unitsX = UNITS_Fraction,		//!< Units in hotspot for X
			UNITS unitsY = UNITS_Fraction			//!< Units in hotspot for Y
			);

		//! Add LinearRing element.
		ERRVALUE LinearRingAdd (
			const POLYLINE& line,					//! Line coordinates (2D longitude,latitude or 3D lon,lat,altitude)
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false,					//!< Whether to extrude the point,
			bool tesselate = true,					//!< Whether to tesselate, note this is opposite the KML default	
			double gx_altitudeOffset = 0			//!< Altitude offset
			);

		//! Add LineString element.
		ERRVALUE LineStringAdd (
			const POLYLINE& line,					//! Line coordinates (2D longitude,latitude or 3D lon,lat,altitude)
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false,					//!< Whether to extrude the point,
			bool tesselate = true,					//!< Whether to tesselate, note this is opposite the KML default	
			double gx_altitudeOffset = 0,			//!< Altitude offset
			int gx_drawOrder = 0						//!< Drawing order
			);

		//! Add "LabelStyle" section.
		ERRVALUE LabelStyleAdd (
			const COLOR& color,						//!< Label color
			double scale = 1							//!< Resizing scale
			);

		//! Add "LineStyle" section.
		ERRVALUE LineStyleAdd (
			const COLOR& color,						//!< Line color
			double width = 1,							//!< Width in pixels
			const COLOR *pgx_outerColor = 0,		//!< Color of portion of line defined by gx:outerWidth
			double gx_outerWidth = 0,				//!< Proportion of line that uses gx:outerColor
			double gx_physicalWidth = 0,			//!< Physical width in meters
			bool gx_labelVisibility = false		//!< Whether to display text label on a LineString
			);

		//! Add link element.
		ERRVALUE LinkAdd (
			const MISTRING& href,					//!< Location of linked item
			REFRESHMODE refreshMode = REFRESHMODE_OnChange,	//!< Time-based refresh mode
			VIEWREFRESHMODE viewRefreshMode = VIEWREFRESHMODE_Never,	//!< How link is refreshed when "camera" changes
			double refreshInterval = 4,			//!< Refresh interval in seconds for REFRESHMODE_OnInterval
			double viewRefreshTime = 4,			//!< Number of seconds to wait after camera stops for VIEWREFRESHMODE_OnStop
			double viewBoundScale = 1				//!< Scaling for BBOX
			);

		//! Add "ListStyle" section.
		ERRVALUE ListStyleAdd (
			LISTITEMTYPE listItemType,				//!< How feature is displayed in list view
			const COLOR *pbgColor = 0 				//!< Background color
			);

		//! Add "LookAt" element.
		ERRVALUE LookAtAdd (
			const DPOINT3D& ptLookAt,				//!< Longitude, Latitude, Altitude of point to look at
			double range,								//!< Distance to point in meters, must be positive
			double heading,							//!< Heading degrees
			double tilt,								//!< Tilt degrees
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround	//!< Altitude mode
			);

		//! Make KML file.
		ERRVALUE MakeFile (
			const FILEPATH& filepath				//!< Path to file
			);

		//! Begin "MultiGeometry" section.
		ERRVALUE MultiGeometryBegin ()
			{ return (AggregateBegin(ITEMTYPE_MultiGeometry)); }

		//! End "MultiGeometry" section.
		ERRVALUE MultiGeometryEnd ()
			{ return (AggregateEnd(ITEMTYPE_MultiGeometry)); }

		//! Add "name" element.
		ERRVALUE NameAdd (
			const MISTRING& name						//!< Name value
			);

		//! Begin "NetworkLink" section.
		ERRVALUE NetworkLinkBegin (
			bool refreshVisibility = false,		//!< Whether to refresh visibility setting each time refreshed
			bool flyToView = false,					//!< Whether to "fly" to view
			bool visibility = true					//!< Whether initially visible or not
			);

		//! End "NetworkLink" section.
		ERRVALUE NetworkLinkEnd ()
			{ return (AggregateEnd(ITEMTYPE_NetworkLink)); }

		//! Begin "PhotoOverlay" section.
		ERRVALUE PhotoOverlayBegin (
			const MISTRING& name,					//!< Name string
			const MISTRING& href,					//!< Location of linked item
			bool visibility = true,					//!< Whether initially visible
			int drawOrder = 0							//!< Drawing Order
			);

		//! End "PhotoOverlay" section.
		ERRVALUE PhotoOverlayEnd ()
			{ return (AggregateEnd(ITEMTYPE_PhotoOverlay)); }

		//! Begin unnamed "Placemark" section.
		ERRVALUE PlacemarkBegin (
			) { return (PlacemarkBegin(MISTRING::GetEmpty(),MISTRING::GetEmpty())); }

		//! Begin "Placemark" section with name.
		ERRVALUE PlacemarkBegin (
			const MISTRING& name,					//!< Name string
			bool visibility = true					//!< Whether initially visible or not
			) { return (PlacemarkBegin(MISTRING::GetEmpty(),name,visibility)); }

		//! Begin "Placemark" section with ID and name.
		ERRVALUE PlacemarkBegin (
			const MISTRING& id,						//!< Placemark ID
			const MISTRING& name,					//!< Name string
			bool visibility = true					//!< Whether initially visible or not
			);

		//! End "Placemark" section.
		ERRVALUE PlacemarkEnd ()
			{ return (AggregateEnd(ITEMTYPE_Placemark)); }

		//! Add (2D) point.
		ERRVALUE PointAdd2D (
			const DPOINT2D& pt,						//!< Point coordinates, longitude, latitude
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false						//!< Whether to extrude the point
			);

		//! Add (3D) point.
		ERRVALUE PointAdd3D (
			const DPOINT3D& pt,						//!< Point coordinates, longitude, latitude, altitude
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false						//!< Whether to extrude the point
			);

		//! Add Polygon element.
		ERRVALUE PolygonAdd (
			const POLYLINE& polygon,				//! Polygon coordinates (2D longitude,latitude or 3D lon,lat,altitude)
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false,					//!< Whether to extrude the point,
			bool tesselate = true					//!< Whether to tesselate, note this is opposite the KML default	
			);

		//! Add Polygon element from SUBREGION2D.
		ERRVALUE PolygonAdd (
			const SUBREGION2D& subregion,			//! Coordinates (2D longitude,latitude or 3D lon,lat,altitude)
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			bool extrude = false,					//!< Whether to extrude the point,
			bool tesselate = true					//!< Whether to tesselate, note this is opposite the KML default	
			);

		//! Add "PolyStyle" section.
		ERRVALUE PolyStyleAdd (
			const COLOR& color,						//!< Fill color
			bool fill = true,							//!< Whether to fill the polygon
			bool outline = true						//!< Whether to outline the polygon (using current LineStyle)
			);

		//! Add "Region" section.
		ERRVALUE RegionAdd (
			const DRECT2D& rectLatLon,				//!< Rectangle, x=longitude (-180 - 180), y=latitude(-90 - 90)
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			double minLodPixels = 0,				//!< Minimum level of detail pixels
			double maxLodPixels = -1,				//!< Maximum level of detail pixels
			double minFadeExtent = 0,				//!< Minimum fade extent
			double maxFadeExtent = 0				//!< Maximum fade extent
			);

		//! Add "Region" section.
		ERRVALUE RegionAdd (
			const DRECT3D& rectLatLonAlt,			//!< Rectangle, x=longitude (-180 - 180), y=latitude(-90 - 90), z=altitude
			ALTITUDEMODE altitudeMode = ALTITUDEMODE_ClampToGround,		//!< Altitude mode
			double minLodPixels = 0,				//!< Minimum level of detail pixels
			double maxLodPixels = -1,				//!< Maximum level of detail pixels
			double minFadeExtent = 0,				//!< Minimum fade extent
			double maxFadeExtent = 0				//!< Maximum fade extent
			);

		//! Begin "Schema" section.
		ERRVALUE SchemaBegin (
			const MISTRING& name						//!< Schema name
			) { return (AggregateBegin(ITEMTYPE_Schema,"name",name)); }

		//! Begin "Schema" section.
		ERRVALUE SchemaBegin (
			const char *name							//!< Schema name
			) { return (AggregateBegin(ITEMTYPE_Schema,"name",MISTRING(name))); }

		//! End "Schema" section.
		ERRVALUE SchemaEnd ()
			{ return (AggregateEnd(ITEMTYPE_Schema)); }

		//! Begin "SchemaData" section.
		ERRVALUE SchemaDataBegin (
			const MISTRING& schemaUrl				//!< Schema Url
			) { return (AggregateBegin(ITEMTYPE_SchemaData,"schemaUrl",schemaUrl)); }

		//! End "SchemaData" section.
		ERRVALUE SchemaDataEnd ()
			{ return (AggregateEnd(ITEMTYPE_SchemaData)); }

		//! Begin "ScreenOverlay" section.
		ERRVALUE ScreenOverlayBegin (
			const MISTRING& href,					//!< Location of linked item
			const VEC2& overlayXY,					//!< Reference point in overlay
			const VEC2& screenXY,					//!< Reference point on screen
			const VEC2 *psizeXY = 0,				//!< Resizing values
			int drawOrder = 0,						//!< Drawing Order
			double rotation = 0,						//!< Rotation angle
			const VEC2 *protationXY = 0			//!< Rotation point
			);

		//! End "ScreenOverlay" section.
		ERRVALUE ScreenOverlayEnd ()
			{ return (AggregateEnd(ITEMTYPE_ScreenOverlay)); }

		//! Set whether to indent for readability.
		//! By default, elements are not indented.
		//! Should only be used before MakeFile is called.
		void SetIndent (
			bool indent									//!< Whether to indent
			) { m_bIndent = indent; }

		//! Set precision for X/Y (longitude/latitude) coordinates.
		//! Default precision is 7 (approximately centimeter)
		void SetPrecisionXY (
			int precisionXY							//!< Precision (number of decimal places)
			) { m_precisionXY = precisionXY; }

		//! Set precision for Z (altitude) coordinates.
		//! Default precision is 2 (centimeter)
		void SetPrecisionZ (
			int precisionZ								//!< Precision (number of decimal places)
			) { m_precisionZ = precisionZ; }

		//! Set whether to use newlines after items in file.
		//! By default, each element is on own line in output.
		//! Should only be used before MakeFile is called.
		void SetSeparateLines (
			bool separate								//!< Whether to separate lines
			) { m_bSeparateLines = separate; }

		//! Add "SimpleData" element.
		ERRVALUE SimpleDataAdd (
			const MISTRING& name,					//!< Name attribute
			const MISTRING& value					//!< Value
			);

		//! Add "SimpleData" element.
		ERRVALUE SimpleDataAdd (
			const char *name,							//!< Name attribute
			const MISTRING& value					//!< Value
			);

		//! Add "SimpleField" element.
		ERRVALUE SimpleFieldAdd (
			SIMPLEFIELDTYPE type,					//!< Type attribute
			const MISTRING& name,					//!< Name attribute
			const MISTRING *pDisplayName = 0		//!< Optional displayName string
			);

		//! Add "SimpleField" element.
		ERRVALUE SimpleFieldAdd (
			SIMPLEFIELDTYPE type,					//!< Type attribute
			const char *name,							//!< Name attribute
			const char *DisplayName = 0			//!< Optional displayName string
			);

		//! Add "Snippet" element.
		ERRVALUE SnippetAdd (
			const MISTRING& text,					//!< Snippet text
			int maxLines = 2							//!< Maximum lines property
			);

		//! Add "Style" or "StyleUrl" using LINESTYLE.
		//! If shared styles have previously been added and SHAREDSTYLEUSE_None is not specified, the set of shared
		//! styles will be searched for a match, and if found a "StyleUrl" item will be added with the appropriate style ID.
		ERRVALUE StyleAdd (
			const LINESTYLE& linestyle,			//!< Line style (only DrawColor and scaling members are used)
			SHAREDSTYLEUSE shareduse = SHAREDSTYLEUSE_IfAvailable		//!< How to use shared styles if any
			);

		//! Add "Style" section using LINESTYLE with ID.
		//! Should only be used within "Document" section.
		ERRVALUE StyleAdd (
			const LINESTYLE& linestyle,			//!< Line style (only DrawColor and scaling members are used)
			const MISTRING& id						//!< Style ID
			);

		//! Add "Style" section using POINTSTYLE.
		//! If shared styles have previously been added and SHAREDSTYLEUSE_None is not specified, the set of shared
		//! styles will be searched for a match, and if found a "StyleUrl" item will be added with the appropriate style ID.
		ERRVALUE StyleAdd (
			const POINTSTYLE& pointstyle,			//!< Style definition
			SHAREDSTYLEUSE shareduse = SHAREDSTYLEUSE_IfAvailable		//!< How to use shared styles if any
			);

		//! Add "Style" section using POINTSTYLE with ID.
		//! Should only be used within "Document" section.
		ERRVALUE StyleAdd (
			const POINTSTYLE& pointstyle,			//!< Style definition
			const MISTRING& id						//!< Style ID
			);

		//! Add "Style" section using POLYSTYLE.
		//! If shared styles have previously been added and SHAREDSTYLEUSE_None is not specified, the set of shared
		//! styles will be searched for a match, and if found a "StyleUrl" item will be added with the appropriate style ID.
		ERRVALUE StyleAdd (
			const POLYSTYLE& polystyle,			//!< Style definition
			SHAREDSTYLEUSE shareduse = SHAREDSTYLEUSE_IfAvailable		//!< How to use shared styles if any
			);

		//! Add "Style" section using POLYSTYLE with ID.
		//! Should only be used within "Document" section.
		ERRVALUE StyleAdd (
			const POLYSTYLE& polystyle,			//!< Style definition
			const MISTRING& id						//!< Style ID
			);

		//! Add "Style" section using TEXTSTYLE.
		//! This will setup for use with Point feature to show name as text.
		//! If shared styles have previously been added and SHAREDSTYLEUSE_None is not specified, the set of shared
		//! styles will be searched for a match, and if found a "StyleUrl" item will be added with the appropriate style ID.
		ERRVALUE StyleAdd (
			const TEXTSTYLE& textstyle,			//!< Text style (only FGColor and scaling members are used)
			SHAREDSTYLEUSE shareduse = SHAREDSTYLEUSE_IfAvailable		//!< How to use shared styles if any
			);

		//! Add "Style" section using TEXTSTYLE with ID.
		//! This will setup for use with Point feature to show name as text.
		//! Should only be used within "Document" section.
		ERRVALUE StyleAdd (
			const TEXTSTYLE& textstyle,			//!< Text style (only FGColor and scaling members are used)
			const MISTRING& id						//!< Style ID
			);

		//! Add shared "Style" section (if unique) from LINESTYLE.
		//! Shared styles may only be added to the "Document" section.
		ERRVALUE StyleAddShared (
			const LINESTYLE& linestyle				//!< Style
			);

		//! Add shared "Style" section (if unique) from POINTSTYLE.
		//! Shared styles may only be added to the "Document" section.
		ERRVALUE StyleAddShared (
			const POINTSTYLE& pointstyle			//!< Style
			);

		//! Add shared "Style" section (if unique) from POLYSTYLE.
		//! Shared styles may only be added to the "Document" section.
		ERRVALUE StyleAddShared (
			const POLYSTYLE& polystyle				//!< Style
			);

		//! Add shared "Style" section (if unique) from TEXTSTYLE.
		//! Shared styles may only be added to the "Document" section.
		ERRVALUE StyleAddShared (
			const TEXTSTYLE& textstyle				//!< Style
			);

		//! Begin "Style" section.
		ERRVALUE StyleBegin ()
			{ return (AggregateBegin(ITEMTYPE_Style)); }

		//! Begin "Style" section with ID.
		ERRVALUE StyleBegin (
			const MISTRING& id						//!< Style ID
			) { return (AggregateBegin(ITEMTYPE_Style,"id",id)); }

		//! End "Style" section.
		ERRVALUE StyleEnd ()
			{ return (AggregateEnd(ITEMTYPE_Style)); }

		//! Add "styleUrl" item.
		ERRVALUE StyleUrlAdd (
			const MISTRING& styleUrl				//!< URI string
			);

		//! Add "TimeSpan" element using DATETIMEs.
		ERRVALUE TimeSpanAdd (
			const DATETIME& dateTimeBegin,		//!< Beginning date/time, none if invalid
			const DATETIME& dateTimeEnd			//!< Ending date/time, none if invalid
			);

		//! Add "TimeSpan" element using MISTRINGs.
		ERRVALUE TimeSpanAdd (
			const MISTRING& begin,					//!< Beginning date/time, empty for none
			const MISTRING& end						//!< Ending date/time, empty for none
			);

		//! Add "TimeStamp" element using DATETIME.
		ERRVALUE TimeStampAdd (
			const DATETIME& when						//!< Date/time
			);

		//! Add "TimeStamp" element using MISTRING.
		ERRVALUE TimeStampAdd (
			const MISTRING& when						//!< Date/time string
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		enum BASETYPE {
			BASETYPE_Object,
			BASETYPE_Feature,
			BASETYPE_Overlay,
			BASETYPE_Container,
			BASETYPE_Geometry,
			BASETYPE_StyleSelector,
			BASETYPE_TimePrimitive,
			BASETYPE_AbstractView,
			BASETYPE_SubStyle,
			BASETYPE_ColorStyle,
			BASETYPE_gxTourPrimitive,
			BASETYPE_Invalid
			};

		enum ITEMTYPE {
			ITEMTYPE_BalloonStyle,
			ITEMTYPE_Camera,
			ITEMTYPE_Data,
			ITEMTYPE_Document,
			ITEMTYPE_ExtendedData,
			ITEMTYPE_Folder,
			ITEMTYPE_GroundOverlay,
			ITEMTYPE_Icon,
			ITEMTYPE_IconStyle,
			ITEMTYPE_Kml,
			ITEMTYPE_LabelStyle,
			ITEMTYPE_LatLonAltBox,
			ITEMTYPE_LatLonBox,
			ITEMTYPE_LinearRing,
			ITEMTYPE_LineString,
			ITEMTYPE_LineStyle,
			ITEMTYPE_Link,
			ITEMTYPE_ListStyle,
			ITEMTYPE_Location,
			ITEMTYPE_Lod,
			ITEMTYPE_LookAt,
			ITEMTYPE_Model,
			ITEMTYPE_MultiGeometry,
			ITEMTYPE_NetworkLink,
			ITEMTYPE_NetworkLinkControl,
			ITEMTYPE_Orientation,
			ITEMTYPE_PhotoOverlay,
			ITEMTYPE_Placemark,
			ITEMTYPE_Point,
			ITEMTYPE_Polygon,
			ITEMTYPE_PolyStyle,
			ITEMTYPE_Region,
			ITEMTYPE_Scale,
			ITEMTYPE_Schema,
			ITEMTYPE_SchemaData,
			ITEMTYPE_ScreenOverlay,
			ITEMTYPE_SimpleData,
			ITEMTYPE_SimpleField,
			ITEMTYPE_Style,
			ITEMTYPE_StyleMap,
			ITEMTYPE_TimeSpan,
			ITEMTYPE_TimeStamp,
			ITEMTYPE_gxAnimatedUpdate,
			ITEMTYPE_gxFlyTo,
			ITEMTYPE_gxLatLonQuad,
			ITEMTYPE_gxMultiTrack,
			ITEMTYPE_gxPlayList,
			ITEMTYPE_gxSoundCue,
			ITEMTYPE_gxTimeSpan,
			ITEMTYPE_gxTimeStamp,
			ITEMTYPE_gxTour,
			ITEMTYPE_gxTourControl,
			ITEMTYPE_gxTrack,
			ITEMTYPE_gxWait,
			ITEMTYPE_coordinates,
			ITEMTYPE_innerBoundaryIs,
			ITEMTYPE_outerBoundaryIs,
			ITEMTYPE_Snippet,
			ITEMTYPE_Invalid
			};

		class SHAREDSTYLEMGR;

		FILE_TEXT m_file;
		MISTRING m_str;
		SIMPLE_ARRAY<ITEMTYPE> m_TypeStack;
		SHAREDSTYLEMGR *m_pSharedStyleMgr;
		int m_precisionXY;
		int m_precisionZ;
		bool m_bIndent;
		bool m_bSeparateLines;

		static void CalcStyleValues (const LINESTYLE& style, double& width);
		static void CalcStyleValues (const POINTSTYLE& style, MISTRING& iconurl, double& scale);
		static void CalcStyleValues (const POLYSTYLE& style, bool& fill, bool& outline, double& width);
		static void CalcStyleValues (const TEXTSTYLE& style, double& scale);
		static BASETYPE GetBaseType (BASETYPE basetype);
		static BASETYPE GetBaseType (ITEMTYPE itemtype);
		static bool HasBaseTypeOf (ITEMTYPE itemtype, BASETYPE basetypetest);
		static const char* ItemTypeStr (ITEMTYPE itemtype);
		static MISTRING PrimitiveBegin (const char *name);
		static MISTRING PrimitiveEnd (const char *name);
		static const char* SimpleFieldTypeStr (SIMPLEFIELDTYPE type);

		void AddLonLatAltMode (const DPOINT3D& pt, ALTITUDEMODE altitudeMode);
		ERRVALUE AddPolyline (const POLYLINE& polyline, bool close = false);
		void AddVec2 (const char *name, double x, double y, UNITS xunits, UNITS yunits);
		void AddVec2 (const char *name, const VEC2& vec2);
		ERRVALUE AggregateBegin (ITEMTYPE itemtype);
		ERRVALUE AggregateBegin (ITEMTYPE itemtype, const char *attributeName, const MISTRING& attributeValue);
		ERRVALUE AggregateEnd (ITEMTYPE itemtype);
		bool CanAddItem (ITEMTYPE itemtype) const;
		ERRVALUE CheckStack (ITEMTYPE itemtype) const;
		ERRVALUE ContainerBegin (ITEMTYPE itemtype, bool open, LISTITEMTYPE listItemType, const COLOR *pListBgColor);
		bool ContainerHasBaseType (BASETYPE basetype) const;
		MISTRING FormatCoord2D (const DPOINT2D& pt) const;
		MISTRING FormatCoord3D (const DPOINT2D& pt, double z) const;
		MISTRING FormatCoord3D (const DPOINT3D& pt) const
			{ return (FormatCoord3D(pt,pt.z)); }
		MISTRING Indent () const;

		WRITER (const WRITER&);
		WRITER& operator= (const WRITER&);
		#endif
	};

//=====================================================================================================================
}	// End namespace MIKML

#endif	// INC_MI32_KMLWRITER_H
