/**
 * \file <mi32/lidarlasdataset.h>
 * \brief LIDAR::LAS::DATASET definitions
 *
 * \if NODOC
 * $Id: lidarlasdataset.h_v 1.14 2013/02/26 16:24:58 mju Exp $
 *
 * $Log: lidarlasdataset.h_v $
 * Revision 1.14  2013/02/26 16:24:58  mju
 * Add getTimeStampLatest.
 *
 * Revision 1.13  2012/10/10 15:27:09  mju
 * Add getClassUsedSet.
 *
 * Revision 1.12  2012/10/08 20:41:18  mju
 * Support access to previous classification.
 *
 * Revision 1.11  2012/09/27 16:14:41  mju
 * Add pointiterator::setSkipWithheld.
 *
 * Revision 1.10  2011/08/08 17:56:15  mju
 * Add scanTotalPoints.
 *
 * Revision 1.9  2011/08/05 21:50:37  mju
 * Add add/removeDelegateOnChangeClassDescSet.
 *
 * Revision 1.8  2011/08/05 17:57:01  mju
 * Add pointDensity value.
 *
 * Revision 1.7  2011/08/04 20:06:42  mju
 * Add pointiterator accessors for non-const and updateFile.
 *
 * Revision 1.6  2011/07/22 14:22:48  mju
 * Add setClassUsed.
 *
 * Revision 1.5  2011/07/21 21:12:06  mju
 * Add setStatIntensity.
 *
 * Revision 1.3  2011/07/18 20:30:23  mju
 * Add scan method and access to stats for intensity.
 *
 * Revision 1.1  2011/07/08 15:45:12  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_LIDARLASDATASET_H
#define	INC_MI32_LIDARLASDATASET_H

#ifndef	INC_MI32_LIDARLASPOINT_H
#include <mi32/lidarlaspoint.h>
#endif

#ifndef	INC_MI32_RECT_H
#include <mi32/rect.h>
#endif

#ifndef	INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif

#ifndef	GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class UNSIGNEDSET;
class BITSET;
class DATETIME;
class FILEPATH;
class FILEPATHLIST;
class REGION2D;
class STAT_UNIVARIATE;
namespace LIDAR {
	class CLASSDESCSET;
	namespace LAS {
		class FILE;
		}
	}
namespace SPATREF {
	class COORDREFSYS;
	}
#endif

namespace LIDAR {
namespace LAS {


//=====================================================================================================================

//! LAS dataset consisting of one or more LAS files along with class descriptions and CRS.
class DATASET {
	public:


		//! Iterator for sequentially accessing point records in dataset.
		class POINTITERATOR {
			public:

				//! Construct with associated dataset.
				explicit POINTITERATOR (
					DATASET& dataset
					);

				~POINTITERATOR ();

				//! Return reference to point record.
				const POINTRECORD& operator* (
					) const { return (m_PointRecord); }

				//! Return reference to non-const point record.
				POINTRECORD& operator* (
					) { return (m_PointRecord); }

				//! Return pointer to point record.
				const POINTRECORD* operator-> (
					) const { return (&m_PointRecord); }

				//! Return pointer to non-const point record.
				POINTRECORD* operator-> (
					) { return (&m_PointRecord); }

				//! Increment to next point, will set to "end" if error occurs.
				POINTITERATOR& operator++ ();

				//! Determine if at end of iteration.
				bool AtEnd () const
					{ return (m_AtEnd); }

				//! Begin iteration.
				ERRVALUE Begin ();

				//! Find closest point using iterator.
				//! Iterator will be at end when returns.
				//! @return Point number within dataset or -1 if none found.
				INT64 FindClosest (
					const DPOINT2D& pt,			//!< Point to search around in dataset CRS
					double SearchDist,			//!< Search distance in dataset CRS units
					double *pDistFound = 0		//!< Distance to point found returned, NULL if don't need
					);

				//! Get "point number" within dataset.
				INT64 GetPointNum () const;

				//! Increment to next point, does not set to "end" if error occurs.
				ERRVALUE Increment ();

				//! Set delegate for reporting progress.
				void SetDelegateOnReportProgress (
					DELEGATE_ERRVALUE_DOUBLE delegate	//!< Delegate to report progress
					);

				//! Set extents for iteration.
				//! Must be done prior to calling Begin().
				void SetExtents (
					const DRECT2D& extents					//!< Extents in dataset CRS
					);

				//! Set filter delegate.
				//! Must be done prior to calling Begin().
				void SetFilter (
					DELEGATE_BOOL_POINTRECORD delegate	//!< Delegate for filtering point records
					);

				//! Set points to be iterated over.
				//! Specified set must remain valid for the life of the iterator.
				//! Must be done prior to calling Begin().
				void SetPointSet (
					const UNSIGNEDSET *pPointSet			//!< Set of points to be iterated over, must remain valid for life of iterator
					);

				//! Set whether to skip "withheld" points.
				void SetSkipWithheld (
					bool SkipWithheld = true				//!< true to skip, false to not
					);

				//! Set to use previous classification values.
				//! Previous class will be read into "user data" field of point records.
				void SetUseClassPrev (
					const DATETIME& TimeStamp				//!< Timestamp, previous classification will at least as old as specified
					);

				//! Update current point record to file.
				ERRVALUE UpdateFile ();

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				class PRIV;
				PRIV *m_pPriv;
				POINTRECORD m_PointRecord;
				bool m_AtEnd;
				#endif
			};

		//! Default constructor.
		DATASET ();

		//! Destructor.
		~DATASET ();

		//! Add delegate for notification of changes to class descriptions.
		//! Changes include setting as "used" and wholesale replacement.
		void AddDelegateOnChangeClassDescSet (
			DELEGATE_VOID_NOPARMS delegate		//!< Delegate to add
			);

		//! Add single LAS file to dataset.
		ERRVALUE AddFile (
			const FILEPATH& laspath
			);

		//! Add list of files to dataset.
		ERRVALUE AddFiles (
			const FILEPATHLIST& laspathlist
			);

		//! Clear dataset.
		void Clear ();

		//! Get class description set.
		const CLASSDESCSET& GetClassDescSet () const;

		//! Get list of available previous classification timestamps.
		ERRVALUE GetClassPrevTimeStampList (
			MILIST<DATETIME>& TimeStampList		//!< List of timestamps returned
			) const;

		//! Get set of classes used in previous classification.
		ERRVALUE GetClassPrevUsedSet (
			const DATETIME& TimeStamp,				//!< Previous classification timestamp
			BITSET& ClassUsedSet						//!< Set of classes used returned
			) const;

		//! Get set of classes used in current classification.
		const BITSET& GetClassUsedSet () const;

		//! Get CRS.
		const SPATREF::COORDREFSYS& GetCRS () const;

		//! Get dataset extents.
		const DRECT3D& GetExtents () const;

		//! Get point density.
		double GetPointDensity () const;

		//! Get dataset region.
		ERRVALUE GetRegion (
			REGION2D& region
			) const;

		//! Get statistics for intensity.
		const STAT_UNIVARIATE& GetStatIntensity () const;

		//! Get timestamp of most recently changed file in dataset.
		ERRVALUE GetTimeStampLatest (
			DATETIME& timestamp				//!< Timestamp returned
			) const;

		//! Get total number of known points in dataset.
		INT64 GetTotalPoints () const;

		//! Determine if total number of points is known.
		bool HasTotalPoints () const;

		//! Determine if class number is used.
		bool IsClassUsed (
			UINT16 ClassNumber
			) const;

		//! Open dataset using definition file.
		ERRVALUE Open (
			const FILEPATH& filepath				//!< Path to dataset definition file 
			);

		//! Read single point.
		ERRVALUE ReadPoint (
			INT64 PointNum,							//!< Point number within dataset to read
			POINTARRAY& PointArray					//!< Point array to read into
			);

		//! Remove delegate for notification of changes to class descriptions.
		void RemoveDelegateOnChangeClassDescSet (
			DELEGATE_VOID_NOPARMS delegate		//!< Delegate to remove
			);

		//! Save dataset definition file.
		ERRVALUE Save (
			const FILEPATH& filepath				//!< Path to dataset definition file
			);

		//! Scan dataset for various usage and statistic values.
		ERRVALUE Scan ();

		//! Scan dataset for total points.
		//! Will open each linked LAS file for which count is not already known.
		ERRVALUE ScanTotalPoints () const;

		//! Set class description set.
		void SetClassDescSet (
			const CLASSDESCSET& classdesc			//!< Class descriptions
			);

		//! Set whether class is used or not.
		void SetClassUsed (
			UINT16 ClassNumber,						//!< Class number
			bool isused = true,						//!< Used
			bool notify = true						//!< Notify via delegates
			);

		//! Set CRS.
		void SetCRS (
			const SPATREF::COORDREFSYS& crs		//!< CRS for dataset
			);

		//! Set intensity statistics.
		void SetStatIntensity (
			const STAT_UNIVARIATE StatIntensity	//!< Intensity statistics to set
			);

		//! Update dataset definition file if changed.
		ERRVALUE Update ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class FILEITEM;
		class PRIV;
		PRIV *m_pPriv;

		DATASET (DATASET&);
		DATASET& operator= (DATASET&);
		#endif
	};

//=====================================================================================================================

}	// End namespace LAS
}	// End namespace LIDAR

#endif	// INC_MI32_LIDARLASDATASET_H
