/**
 * \file serializer.h
 * \brief SERIALIZER class definition.
 *
 * \if NODOC
 * $Id: serializer.h_v 1.12 2012/06/11 19:56:40 bcolombe Exp $
 *
 * $Log: serializer.h_v $
 * Revision 1.12  2012/06/11 19:56:40  bcolombe
 * Support reading/writing to string
 *
 * Revision 1.11  2012/06/07 15:31:44  mju
 * Add putBool and putItem with rvc::objitem.
 *
 * Revision 1.10  2011/01/18 17:19:32  mju
 * Add rvc::object::autofilecache member.
 *
 * Revision 1.9  2010/06/03 14:40:00  mju
 * Make m_parmvalue mistring so not limited in size.
 *
 * Revision 1.8  2007/11/12 16:22:19  mju
 * Use mistatuscontext.
 *
 * Revision 1.7  2007/09/06 20:57:35  mju
 * Add putItem for double_range.
 *
 * Revision 1.6  2007/08/29 20:32:19  mju
 * Add append method.
 *
 * Revision 1.5  2007/05/08 13:02:24  mju
 * Use objitemlist hdr.
 *
 * Revision 1.4  2007/01/08 15:06:19  mju
 * Remove unused fns for subObjects.
 *
 * Revision 1.3  2006/10/05 14:18:31  mju
 * Use filefind to search for related file.
 *
 * Revision 1.2  2006/09/15 16:46:02  mju
 * Remove deprecated defns as no longer used.
 *
 * Revision 1.1  2006/09/15 15:24:23  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_SERIALIZER_H
#define	INC_MI32_SERIALIZER_H

#ifndef	INC_MI32_SERIALIZERITEM_H
#include <mi32/serializeritem.h>
#endif

#ifndef	INC_RVC_OBJITEMLIST_H
#include <rvc/objitemlist.h>
#endif

#ifndef	INC_RVC_SERIALIZE_H
#include <rvc/serialze.h>
#endif

#ifndef	INC_MI32_RANGE_H
#include <mi32/range.h>
#endif

#ifdef WIN32
#pragma warning (disable:4996)	// Disable warning about deprecated declaration within header
#endif

// Forward declarations
#ifndef GENERATING_DOXYGEN_OUTPUT
class FILE_TEXT;
struct OLDPROJPARM;
struct COLOR;
struct DPOINT2D;
struct DPOINT3D;
struct DRECT2D;
struct DRECT3D;
class MISTATUSCONTEXT;
namespace SPATREF {
	class COORDREFSYS;
	}
#endif

//=====================================================================================================================
//! Class for serializing object instances to file.
//! The SERIALIZER class provides methods to write object instances to a text
//! file/object and read those instances back into memory later.  The design
//! allows for the object structure to change while maintaining compatibility
//! with previously existing files.

class SERIALIZER {
	public:

		//! Default constructor.
		SERIALIZER (
			);

		//! Destructor.
		virtual ~SERIALIZER (
			);

		//! Open text file containing serialization to append.
		ERRVALUE Append (
			const FILEPATH& filepath			//!< Text file path
			);

		//! Close the serializer, will be done in destructor.
		void Close (
			);

		//! Get path to serialization file.
		const FILEPATH& GetFilePath (
			) const { return (m_filepath); }

		//! Determine if ignoring "skippable" items.
		bool GetIgnoreSkippable (
			) const { return (m_IgnoreSkippable); }

		//! Get name of current item being read/written.
		const char* GetItemName (
			) const { return (m_itemname); }

		//! Get value for current item being read.
		const MIUNICODE* GetItemValue (
			) const { return (m_parmvalue); }

		//! Get value for current item as double.
		//! @return true if could convert, false if not
		bool GetItemValue (
			double& value							//!< Value returned
			) const;

		// Get log text.
		const MISTRING& GetLog (
			) const { return (m_Log); }

		// Get log text (non-const).
		MISTRING& GetLog (
			) { return (m_Log); }

		//! Get name of object, will be blank if text file.
		const MIUNICODE* GetObjectName (
			) const { return (m_Descriptor.GetFullName()); }

		//! Get list of missing files/objects after scan, also clears the list.
		void GetScanListMissing (
			RVC::OBJITEMLIST& MissingList
			) { MissingList.Merge(m_ScanFileObjListMissing); }

		//! Get list of "used" files/objects after scan, also clears the list.
		void GetScanListUsed (
			RVC::OBJITEMLIST& UsedList
			) { UsedList.Merge(m_ScanFileObjListUsed); }

		//! Get serialized string.  If no filepath or object opened, will output to string by default
		const MISTRING& GetText (
			) const { return (m_serialtext); }

		//! Set serialized string.  If no filepath or object opened, will read from string by default
		void SetText (
			const MISTRING& serialtext
			) { m_serialtext = serialtext; }

		//! Get current version setting.
		UINT16 GetVersion (
			) const { return (m_Version[m_depth]); }

		//! Check if scanning serialization only.
		bool IsScanOnly (
			) const { return (m_ScanOnly); }

		//! Clear serialization log.
		void LogClear (
			) { m_Log.Clear(); }

		//! Add error message to log.
		//! Will insert newline before and after error text.
		void LogError (
			ERRVALUE err
			);

		//! Add text to log.
		void LogText (
			const MISTRING& text
			) { m_Log << text; }

		//! Create object to contain serialization using OBJITEM.
		ERRVALUE Make (
			RVC::OBJITEM& objitem,				//!< Object item reference passed/returned
			const RVC::SERIALIZE::MAKEPARMS& makeparms
			);

		//! Create object to contain serialization using parent and descriptor.
		ERRVALUE Make (
			RVC::OBJECT& parent,					//!< Parent object
			RVC::OBJTYPE objtype,				//!< Object type
			RVC::DESCRIPTOR& descriptor,		//!< Descriptor for new object
			const RVC::SERIALIZE::MAKEPARMS& makeparms
			);

		//! Create text file to contain serialization.
		ERRVALUE Make (
			const FILEPATH& filepath,			//!< Text file path
			int version = 1						//!< Version to write
			);

		//! Open object containing serialization to read.
		ERRVALUE Open (
			const RVC::OBJITEM& objitem		//!< RVC object item reference
			);

		//! Open text file containing serialization to read.
		ERRVALUE Open (
			const FILEPATH& filepath			//!< Text file path
			);

		//! Write beginning tag for starting a component.
		ERRVALUE PutBegin (
			const char *tagname,					//!< Tag name to write
			UINT16 version = 0					//!< Optional version
			);

		//! Write bool value.
		ERRVALUE PutBool (
			const char *tagname,					//!< Tag name
			bool value								//!< Value to write
			);

		//! Write floating-point value.
		ERRVALUE PutDouble (
			const char *tagname,					//!< Tag name
			double value							//!< Value to write
			);

		//! Write ending tag at end of a component.
		ERRVALUE PutEnd ();

		//! Write integer value.
		ERRVALUE PutInt (
			const char *tagname,					//!< Tag name to write
			int value								//!< Value to write
			);

		//! Write COLOR item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const COLOR& color					//!< Item value
			);

		//! Write DPOINT2D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const DPOINT2D& point2d				//!< Item value
			);

		//! Write DPOINT3D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const DPOINT3D& point3d				//!< Item value
			);

		//! Write DRECT2D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const DRECT2D& rect					//!< Item value
			);

		//! Write DRECT3D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const DRECT3D& rect					//!< Item value
			);

		//! Write LPOINT2D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const LPOINT2D& point2d				//!< Item value
			);

		//! Write LRECT2D item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const LRECT2D& rect					//!< Item value
			);

		//! Write DOUBLE_RANGE item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const DOUBLE_RANGE& value			//!< Item value
			);

		//! Write OLDPROJPARM item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const OLDPROJPARM& oldprojparm	//!< Item value
			);

		//! Write SPATREF::COORDREFSYS item.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const SPATREF::COORDREFSYS& crs	//!< Item value
			);

		//! Write from OBJITEM.
		ERRVALUE PutItem (
			const char *tagname,					//!< Tag name to write
			const RVC::OBJITEM& objitem,		//!< Item value
			SERIALIZERITEM::TYPE type			//!< Item type, must be one of TYPE_RVCOBJITEM... or error will result
			);

		//! Write a single-line Unicode string
		ERRVALUE PutStr (
			const char *tagname,					//!< Tag name to write
			const MIUNICODE *str					//!< String
			);

		//! Write a Multi-line Unicode String
		ERRVALUE PutStrML (
			const char *tagname,					//!< Tag name to write
			const MIUNICODE *str					//!< String
			);

		//! Write unsigned integer value.
		ERRVALUE PutUnsigned (
			const char *tagname,					//!< Tag name to write
			unsigned int value					//!< Item value
			);

		//! Read SERIALIZABLE object.
		ERRVALUE Read (
			SERIALIZABLE& object					//!< Object to read
			);

		//! Read SERIALIZABLE object.
		ERRVALUE Read (
			const char *tagname,					//!< Tag name to match
			SERIALIZABLE& object					//!< Object to read
			);

		//! Read items from serialization.
		ERRVALUE Read (
			const SERIALIZERITEM *itemdef,	//!< Item definition array
			void *data								//!< Serialization data
			);

		//! Read items from serialization.
		ERRVALUE Read (
			const char *tagname,					//!< Tag name to match
			const SERIALIZERITEM *itemdef,	//!< Item definition array
			void *data								//!< Serialization data
			);

		//! Read OLDPROJPARM from serialization.
		ERRVALUE Read (
			OLDPROJPARM& oldprojparm
			);

		//! Rewind file to beginning.
		ERRVALUE Rewind (
			);

		//! Set whether to ignore "skippable" items.
		void SetIgnoreSkippable (
			bool IgnoreSkippable = true		//!< Ignore skippable mode
			) { m_IgnoreSkippable = IgnoreSkippable; }

		//! Set scan-only mode.
		void SetScanOnly (
			bool ScanOnly							//!< Scan only mode
			) { m_ScanOnly = ScanOnly; }

		//! Set version (for overriding).
		void SetVersion (
			UINT16 Version							//!< Version
			) { m_Version[m_depth] = Version; }

		//! Skip remainder of current structure (for use in item callback).
		ERRVALUE SkipStruct (
			);

		//! Update status text from Unicode string.
		ERRVALUE StatusTextUpdate (
			const MISTRING& string
			);

		//! Write items without begin/end tags.
		ERRVALUE Write (
			const SERIALIZERITEM *itemdef,	//!< Item definition array
			const void *data						//!< Data to write
			);

		//! Write items to serialized output.
		ERRVALUE Write (
			const char *tagname,					//!< Name to tag object with
			const SERIALIZERITEM *itemdef,	//!< Item definition array
			const void *data,						//!< Data to write
			UINT16 version = 0					//!< Version to set if top level object
			);

		//! Write SERIALIZABLE object.
		ERRVALUE Write (
			const SERIALIZABLE& object,		//!< Object to write
			const char *tagname = 0				//!< Name to tag object with
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		MISTRING m_serialtext;
		FILEPATH m_filepath;						// Path to file
		RVC::DESCRIPTOR m_Descriptor;			// Descriptor to open RVC text object
		FILE_TEXT *m_pFileText;					// Text file instance if non-RVC
		RVC::SERIALIZE m_SerialObj;			// RVC Serialize instance if Text object
		int m_depth;								// Current depth
		UINT16 m_Version[128];					// Version at current depth
		char m_itemname[128];
		MISTRING m_parmvalue;					// Last parameter value
		MISTATUSCONTEXT *m_pStatusContext;
		INT32 m_filesize;							// Size of file/object being read for status update
		INT32 m_filepos;							// Position in file/object for status update
		bool m_IgnoreSkippable;
		bool m_ScanOnly;
		RVC::OBJITEMLIST m_ScanFileObjListUsed;
		RVC::OBJITEMLIST m_ScanFileObjListMissing;
		MISTRING m_Log;
		RVC::OBJECT::AUTOFILECACHE m_AutoFileCache;

		static const SERIALIZERITEM* GetItemDefREGION2D ();
		static const SERIALIZERITEM* GetItemDefSUBREGION2D ();

		// Call callback for item if necessary.
		ERRVALUE CallItemCB (
			const SERIALIZERITEM *itemdef,
			void *data,								// Callback data
			SERIALIZERITEM::ACTION action				// Callback action
			);

		// Find link file, add to missing file list if can't find it.
		ERRVALUE FindLinkFile (
			const SERIALIZERITEM *parmdp,
			const UINT8* data,
			FILEPATH& linkfilepath,
			bool OriginalFirst = false
			);

		// Find object.
		int FindObject (
			const SERIALIZERITEM *parmdp,
			const SERIALIZERITEM *startdp,
			UINT8 *data
			);

		// Get multi-line string (Unicode).
		ERRVALUE GetStrML (
			MIUNICODE **str							// XXX change to MISTRING&
			);

		// Get multi-line string (ASCII or RAW).
		ERRVALUE GetStrML (
			char **str,
			bool raw = false	
			);

		// 
		ERRVALUE PutStr (
			const char *tagname,
			const char *str
			);

		// 
		ERRVALUE PutStrML (
			const char *tagname,
			const char *str
			);

		// 
		ERRVALUE ReadLine (
			MISTRING& string,
			bool raw = false
			);

		// Write single-line MIUNICODE string.
		ERRVALUE WriteText (
			const MIUNICODE *str					// String to write
			);

		// Write ASCII string.
		ERRVALUE WriteText (
			const char *str						// String to write
			);

		SERIALIZER (const SERIALIZER&);
		SERIALIZER& operator= (const SERIALIZER&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


//=====================================================================================================================

#ifdef WIN32
#pragma warning (default:4996)	// Restore warning about deprecated declaration
#endif

#endif	// INC_MI32_SERIALIZER_H
