/******************************************************************************
 *
 * \file sml/object.h
 * \brief SML OBJECT
 *
 * \if NODOC
 * $Log: object.h_v $
 * Revision 1.9  2012/12/05 17:53:43  mju
 * Add setCurRowColumn.
 *
 * Revision 1.8  2012/11/08 20:33:24  mju
 * Use spatref::coordtrans and just forward-declare.
 *
 * Revision 1.7  2011/10/12 14:31:15  mju
 * Remove redundant hdrs.
 *
 * Revision 1.6  2007/12/31 17:41:00  bcolombe
 * Added COORDOP_LIST to SML::OBJECT for use in ObjectToMap
 *
 * Revision 1.5  2007/09/06 15:54:04  scowan
 * Fixed some comments.
 *
 * Revision 1.4  2007/05/10 21:57:42  dwilliss
 * Added AddRef/Release
 *
 * Revision 1.3  2007/05/10 20:46:36  dwilliss
 * removed some unused methods and members
 *
 * Revision 1.2  2007/05/08 15:46:07  dwilliss
 * SML namespace
 *
 * Revision 1.1  2007/05/04 20:44:14  dwilliss
 * Initial revision
 *
 * \endif
*/
/*****************************************************************************/

#ifndef INC_SML_OBJECT_H
#define INC_SML_OBJECT_H

#ifndef	INC_SML_STYPE_H
	#include <sml/stype.h>
#endif

#ifndef INC_RVC_OBJITEM_H
	#include <rvc/objitem.h>
#endif

#ifndef	INC_MI32_ELEMTYPE_H
	#include <mi32/elemtype.h>
#endif


#ifdef RVCSYSDLL
	#define RVCSYSLIBEXPORT MI_DLLEXPORT
	#define RVCSYSLIBCLASSEXPORT MI_DLLCLASSEXPORT
#else
	#define RVCSYSLIBEXPORT MI_DLLIMPORT
	#define RVCSYSLIBCLASSEXPORT MI_DLLCLASSIMPORT
#endif

#ifndef GENERATING_DOXYGEN_OUTPUT
namespace RVC {
	class FILEHANDLE;
	class OBJECTOBSERVER;
	class OBJECT;
	}
namespace SPATREF {
	class COORDTRANS;
	}
#endif

namespace SML {
class CONTEXT;
class ELEMDATA;
class GEOSTRUCT;
class OBJPROXY;

/*****************************************************************************/
/**
*
* \note Constructor/Destructor are private and only used by SML::CONTEXT
* which is a friend class
*
*/
class RVCSYSLIBCLASSEXPORT OBJECT {
	public:

		void AddRef (
			);

		// Allocate an SML::OBJECT base class.  This is used to hold object
		// information (name, desc, etc) for objects we don't use have any
		// derived class for.  The caller is responsibile for makeing sure
		// the returned pointer is freed some time.
		static OBJECT* Allocate (
			RVC::OBJTYPE objtype
			);

		//! If the object refers to an existing object, delete it, but keep
		//! the objitem setup to point to a NewRequest for the same descriptor
		//! same parent, etc...
		ERRVALUE DeleteExisting (
			);

		//============================ Get Methods ========================
		int GetCurElemNum (
			ELEMTYPE elemtype
			) const;

		int GetElemData (
			ELEMDATA** pElemData,
			ELEMTYPE ElemType,
			bool bAllocIfNeeded = false
			);	//! Not const.  May need to alloc

		//! Get the filepath of the object.
		const FILEPATH& GetFilePath (
			) const;


		//! Returns the RVC::OBJECT for the object.
		//! The derived classes each have a GetInterface which returns the specific
		//! object type. For example  SML::VECTOROBJECT has a GetInterface() that
		//! returns an RVC::VECTOR*
		//! The SML::OBJECT "owns" the pointer returned and will free it when the
		//! object is freed.
		RVC::OBJECT* GetInterfaceBase (
			) const;

		RVC::OBJITEM& GetObjItem (
			);

		const RVC::OBJITEM& GetObjItem (
			) const;

		RVC::OBJTYPE GetObjType (
			) const;

		STYPE GetSMLSTYPE (
			) const;

		bool IsDirty (
			) const;

		void Release (
			);

		//============================ Set Methods ========================

		//! Set the dirty flag so Flush() knows it has to do something
		void SetDirty (
			bool bDirty = true
			);

		//! Set the filepath of the object
		void SetFilePath (
			const FILEPATH& filepath
			);

		//! Pass true to tell SML that the object is a temporary object created
		//! by the Script (eg: by calling CreateTempRaster()).  The object and
		//! its file will be deleted when the script exits.
		void SetIsUserTempObject (
			bool state
			);

		//! Set the RVC::OBJITEM for the object
		void SetObjItem (
			const RVC::OBJITEM& item
			) ;

		//! Setup the object
		//! This is to replace the need to call SetObjItem, SetCredentials, Open
		//! and Close just to set everything up.
		ERRVALUE Setup (
			const RVC::OBJITEM& item,
			const RVC::CREDENTIALS* Credentials
			) ;

		// ========================= Other methods ===========================
		//! Close the object.  Note!  Open/Close calls are not refcounted.  Only
		//! internal SML should call Close() for now.
		//! This is mainly for effeciency.  If you always balanced Open/Close
		//! calls, you'd be opening and closing the object all the time which
		//! would slow down SML.  If you need something open, just call Open() and
		//! let SML close it when it decides it needs to.
		void Close (
			bool CloseDatabasesToo = false
			);

		//! Close the databases associated with an object
		void CloseDatabases (
			);

		//! Not const.  May read last used georef for object if not already read
		int ConvertPointMapToObject (
			const DPOINT2D& src,
			const GEOSTRUCT* SrcGeoref,
			DPOINT2D* pDest
			);

		//! Not const.  May read last used georef for object if not already read
		int ConvertPointObjectToMap (
			const DPOINT2D& src,
			const GEOSTRUCT* DestGeoref,
			DPOINT2D* pDest
			);

		//! Destroies the SML::OBJECT.  "delete this"
		//! Necessary since destructor is private and not exported from DLL
		void Free (
			);

		const RVC::CREDENTIALS* GetCredentials (
			) const;

		//! Returns the "next" object in the context's object list.
		//! Used internally to iterate through the objects.
		OBJECT* Next (
			) const;

	#ifndef GENERATING_DOXYGEN_OUTPUT
	#if defined(RVCSYSDLL) 
		//! Calls the private virtual method v_OpenLow()
		//! NOTE!  Open/Close are not refcounted.  See comment on Close() method.
		//! The Open method used to return an object handle (numeric).  It no
		//! longer does that.  The only thing you should assume about the return
		//! value now is that it will be < 0 for an error.
		//! \deprecated Call GetInterface() and Open that.
		// XXX Still used in qqcode.c
		DEPRECATED_MSG("Call GetInterface() and Open that") 
		int Open (
			int mode		//!< RVC::OBJECTACCESS_Xxxx
			);
	#endif
	#endif


		//! Returns TRUE if SML is busy executing a FOREACH loop on this object
		bool IsDoingForEach (
			ELEMTYPE ElemType
			) const;

		//! Returns TRUE if the object is a temporary object created	by the Script
		//! (eg: by calling CreateTempRaster()).  The object and its file will be
		//! deleted when the script exits.
		bool IsUserTempObject (
			) const;

		//! Read the georef subobject.  If already read, it will just return
		//! what's been read in regardless of inode (I may change that. If
		//! you want to force the matter, you could SetGeoref(NULL) but that
		//! would cause a memory leak).
		//! If inode is 0 (or omitted) it will just read the LastUsed georef
		//! This is a slightly older version of the function which didn't allow
		//! for the possibility of not passing the ObjItem at all to force it
		//! to read the default georeference.
		DEPRECATED int ReadGeoref (
			GEOSTRUCT **pGeostruct,
			const RVC::OBJITEM& ObjItem
			);

		//! Read the georef subobject.  If already read, it will just return
		//! what's been read in regardless of inode (I may change that. If
		//! you want to force the matter, you could SetGeoref(NULL) but that
		//! would cause a memory leak).
		//! If ObjItem is 0 (or omitted) it will just read the LastUsed georef
		int ReadGeoref (
			GEOSTRUCT **pGeostruct,
			const RVC::OBJITEM* ObjItem = 0
			);

		void OnInterfaceDestroy (
			);

		void OnInterfaceAttach (
			const RVC::OBJITEM& ObjItem
			);

		void OnInterfaceDetach (
			);

	#ifndef GENERATING_DOXYGEN_OUTPUT
		static void SetRVCInterfaceProxy (
			OBJPROXY* proxy
			);

		void SetCredentials (
			const RVC::CREDENTIALS* credentials
			);

	#ifdef RVCSYSDLL	// These methods are public but only for use inside RVCSYS

		//! Open the file the object is in.
		//! \deprecated Call GetInterface() and Open that.
		ERRVALUE OpenFileReal (
			int mode		//!< RVC::OBJECTACCESS_Xxxx
			);

		RVCINODENUM GetInodeReal (
			) const;

		ERRVALUE OpenReal (
			int mode,
			const RVC::CREDENTIALS* Credentials = 0
			);

		void SetCurRowColumn (
			INT32 row,
			INT32 col
			) { m_ElemNum[0] = row; m_ElemNum[1] = col; }
	#endif
	#endif

	protected:

		static OBJPROXY* s_Proxy;
		mutable RVC::OBJECT* m_Interface;		// Get method may create one and set it the first time it's called

		void ConstructInterface (
			RVC::OBJTYPE
			) const;

		// Constructor
		OBJECT (
			STYPE type,
			CONTEXT* context
			);

		// Destructor
		virtual ~OBJECT (
			);

	#ifndef GENERATING_DOXYGEN_OUTPUT

		virtual int v_OpenLow (
			int mode
			);

		virtual void v_CloseLow (
			);

		virtual RVC::OBJTYPE v_GetObjType (
			) const;

	#ifdef RVCSYSDLL
	//! For now, allow unrestricted access within rvcapp.dll
	protected:
	#else
	private:
		//! Outside the DLL, none of this should be accessed except through
		//! methods.
	#endif
		INT32 m_RefCount;

		OBJECT *m_prev;
		OBJECT *m_next;
		bool m_SystemTempObject;		//!< created by SML - not user
		bool m_UserTempObject;		//!< explicitly created by user
		//bool m_RegionIsObject;		//!< region was created with GetInputRegion()
									//! or GetOutputRegion()
		bool m_Dirty;					//!< needs to be flushed when closed - regions only for now
		bool m_bIsOpen;				//!< Object is open
		//bool m_FreeInDestructor;	//!< Not in the context's object list
		RVC::OBJITEM m_ObjItem;
	   STYPE m_stype;
	   bool m_bOpenForWrite;
		GEOSTRUCT *m_georef;
		ELEMDATA *m_db[ELEMTYPE_NumTypes];			//!<  pointers to database info
		UINT8 m_DoingForEach[ELEMTYPE_NumTypes];	//!<  ELEMTYPE_RastCell for rasters
		INT32  m_ElemNum[ELEMTYPE_NumTypes];		//!<  use 0 and 1 for lin/col on raster foreach
		CONTEXT *m_context;
		RVC::FILEHANDLE *m_fhandle;
		RVC::OBJECTOBSERVER *m_Observer;
		RVC::CREDENTIALS *m_Credentials;
		SPATREF::COORDTRANS *m_pCoordTrans;

		friend class CONTEXT;		//!< Only SML::CONTEXT can call our ctor/dtor
		friend class VM;
		friend class SYMBOL;
	#endif // GENERATING_DOXYGEN_OUTPUT

   };
} // end of SML namespace

typedef SML::OBJECT SMLOBJECT;


#undef RVCSYSLIBEXPORT
#undef RVCSYSLIBCLASSEXPORT

#endif		//  INC_SML_OBJECT_H


