/**
 * \file spatmod/imagefilterclassify.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_CLASSIFY_MINDISTANCE definition.
 *
 * \if NODOC
 * $Id: imagefilterclassifymindistance.h_v 1.3 2013/01/07 18:24:30 mju Exp $
 *
 * $Log: imagefilterclassifymindistance.h_v $
 * Revision 1.3  2013/01/07 18:24:30  mju
 * Support training from samples.
 *
 * Revision 1.2  2013/01/04 21:33:34  mju
 * Add protected ctor and setClassCenters.
 *
 * Revision 1.1  2013/01/02 17:34:18  mju
 * Initial revision
 *
 * Revision 1.2  2013/01/02 15:42:12  mju
 * Specify min/max classnum in ctor, add setCalcStats, setValueUnclassified.
 *
 * Revision 1.1  2012/12/31 18:52:47  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERCLASSIFYMINDISTANCE_H
#define	INC_SPATMOD_IMAGEFILTERCLASSIFYMINDISTANCE_H

#ifndef	INC_SPATMOD_IMAGEFILTERCLASSIFY_H
#include <spatmod/imagefilterclassify.h>
#endif

#ifndef	INC_MI32_MIMATRIX_H
#include <mi32/mimatrix.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
namespace SPATMOD {
	namespace IMAGE {
		class PIXELTABLE;
		}
	}
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Classify based on minimum "distance" to center.
//! Supports training via samples only.
class FILTER_CLASSIFY_MINDISTANCE : public FILTER_CLASSIFY {
	public:

		//! Distance measurement methods.
		enum MEASURE {
			MEASURE_Euclidean = 0,				//!< Euclidian distance in N-space
			MEASURE_Orthogonal = 1				//!< Sum of distances for each sample
			};

		//! Constructor.
		FILTER_CLASSIFY_MINDISTANCE (
			STAGE *pStageInput,						//!< Pointer to stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			MEASURE measure,							//!< Distance measure
			UINT32 MinClassNum						//!< Minimum class number
			);

		//! Destructor.
		virtual ~FILTER_CLASSIFY_MINDISTANCE ();

		//! Get class centers.
		const MIMATRIX& GetClassCenters () const
			{ return (m_ClassCenters); }

		//! Set class centers using MIMATRIX.
		ERRVALUE SetClassCenters (
			const MIMATRIX& centers,				//!< Class centers
			const BITSET& usedset					//!< Set of classes used
			);

		//! Set class centers using PIXELTABLE.
		ERRVALUE SetClassCenters (
			const PIXELTABLE& centers,				//!< Table of class centers
			const BITSET& usedset					//!< Set of classes used
			);

		//! Set maximum distance, if actual distance is greater, result will be set to "unclassified".
		void SetMaxDistance (
			double MaxDistance
			) { m_MaxDistance = MaxDistance; m_MaxDistance2 = MaxDistance * MaxDistance; }

	protected:

		//! Access class centers.
		MIMATRIX& AccessClassCenters ()
			{ return (m_ClassCenters); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef FILTER_CLASSIFY BASECLASS;

		MIMATRIX m_ClassCenters;
		double m_MaxDistance;
		double m_MaxDistance2;
		MEASURE m_measure;
		SIMPLE_ARRAY<double*> m_SampleMeans;
		SIMPLE_ARRAY<UINT64> m_SampleCounts;

		UINT32 FindClass (UINT8 const * const * pSrcValue) const;
		UINT32 FindClass (double const * const * pSrcValue) const;
		UINT32 FindClass (const SAMPLEITERATOR_CONST *itSrcValue) const;

		// FILTER_CLASSIFY overrides.
		virtual ERRVALUE v_Classify (SAMPLEITERATOR_CONST *itSrcValue, void const **pvSrcValue, UINT8 *pTgtValidity, UINT32 *pTgtValue, UINT32 NumPixels) const;
		virtual ERRVALUE v_TrainFromSamples (SAMPLEITERATOR_CONST *itSrcValue, void const **pSrcValue, const UINT8 *pSrcValidity, const UINT32 *pSampleValue, const UINT8 *pSampleValidity, UINT32 NumPixels);
		virtual ERRVALUE v_TrainPassBegin (bool bAutomatic);
		virtual int v_TrainPassEnd ();

		// Unimplemented
		FILTER_CLASSIFY_MINDISTANCE (const FILTER_CLASSIFY_MINDISTANCE&);
		FILTER_CLASSIFY_MINDISTANCE& operator= (const FILTER_CLASSIFY_MINDISTANCE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERCLASSIFYMINDISTANCE_H
