/**
 * \file imagefilterfocal.h <spatmod/imagefilterfocal.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_FOCAL definition.
 *
 * \if NODOC
 * $Id: imagefilterfocal.h_v 1.7 2011/01/24 18:47:51 mju Exp $
 *
 * $Log: imagefilterfocal.h_v $
 * Revision 1.7  2011/01/24 18:47:51  mju
 * Reorganize so common code in filter_focal to avoid duplication.
 *
 * Revision 1.6  2009/06/23 21:20:32  mju
 * Add setTestWeightGreaterThanOne.
 *
 * Revision 1.4  2007/12/19 22:44:44  bcolombe
 * Add flag to MULTIPLICATIVE to replace null values with center
 * Otherwise always divide by total of valid kernel weights
 *
 * Revision 1.3  2007/12/11 15:33:22  bcolombe
 * Change constructors, will use ComputeSampleStats to get mean and stddev of image
 *
 * Revision 1.2  2007/11/07 23:29:42  bcolombe
 * Add optional amount parameter
 *
 * Revision 1.1  2007/08/30 17:19:53  bcolombe
 * Initial revision
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERFOCAL_H
#define	INC_SPATMOD_IMAGEFILTERFOCAL_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

#ifndef	INC_SPATMOD_IMAGEBUFFER_H
#include <spatmod/imagebuffer.h>
#endif

#ifndef	INC_MI32_MIMATRIX_H
#include <mi32/mimatrix.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Base class for focal filters.
class FILTER_FOCAL : public FILTER {
	public:

		//! Destructor.
		virtual ~FILTER_FOCAL ();

		//! Get amount of output coming from filter (0-1).
		double GetAmountFilter () const
			{ return (m_AmountFilter); }

		//! Get amount of source coming from filter (0-1).
		double GetAmountSource () const
			{ return (m_AmountSource); }

		//! Return number of cells in focal window
		UINT32 GetNumFocalCells () const
			{ return (m_NumFocalCells); }

		//! Return number of columns in focal window
		UINT32 GetNumFocalColumns () const
			{ return (m_NumFocalColumns); }

		//! Return number of rows in focal window
		UINT32 GetNumFocalRows () const
			{ return (m_NumFocalRows); }

		//! Determine whether a binary mask is used to define filtered areas
		bool IsUsingMask () const
			{ return (m_UseMask); }

		//! Set output datatype.
		//! If not used, output type will usually be the same as the input.
		//! If used, must be called before Initialize.
		void SetOutputDataType (DATATYPE DataType)
			{ m_DataType = DataType; }

	protected:

		#ifndef GENERATING_DOXYGEN_OUTPUT
		// Internal.
		class WORKITERATOR;
		#endif

		//! Constructor.
		FILTER_FOCAL (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			UINT32 NumFocalRows,						//!< Number of rows in focal window
			UINT32 NumFocalColumns,					//!< Number of columns in focal window
			double AmountFilterPct = 100			//!< Percentage of output from filter
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			UINT32 NumFocalRows,						//!< Number of rows in focal window
			UINT32 NumFocalColumns,					//!< Number of columns in focal window
			double AmountFilterPct = 100			//!< Percentage of output from filter
			);

		ERRVALUE PullSubBuffers (const SECTION& SubSection, BUFFER& BufSrc, BUFFER& BufMask);

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		double m_AmountFilter;
		double m_AmountSource;
		UINT32 m_NumFocalRows;
		UINT32 m_NumFocalColumns;
		UINT32 m_NumFocalCells;
		DATATYPE m_DataType;
		bool m_UseMask;

		// Unimplemented
		FILTER_FOCAL (const FILTER_FOCAL&);
		FILTER_FOCAL& operator= (const FILTER_FOCAL&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to apply a single kernel on samples in image.  Calculates a weighted average of the cells in the current filter window.
class FILTER_FOCAL_MULTIPLICATIVE : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_MULTIPLICATIVE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image,
			const MIMATRIX& matrix,					//!< Set kernel from matrix
			double AmountFilterPct = 100.0,		//!< Amount of filter to apply (percentage, 0-100)
			bool UseCenterForNull = false			//!< If null value found, use center value for calculations, otherwise skip it
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_MULTIPLICATIVE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed,
			const MIMATRIX& matrix,					//!< Set kernel from matrix
			double AmountFilterPct = 100.0,		//!< Amount of filter to apply (percentage, 0-100)
			bool UseCenterForNull = false			//!< If null value found, use center value for calculations, otherwise skip it
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_MULTIPLICATIVE ();

		//! Get matrix kernel
		const MIMATRIX& GetMatrix (
			) const { return m_Matrix; }

		//! Use matrix to define kernel
		void SetMatrix (
			const MIMATRIX& matrix
			) { m_Matrix = matrix; }

		void SetTestWeightGreaterThanOne (
			) { m_TestWeightGreaterThanOne = true; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		bool m_DoReplaceCenter;
		bool m_TestWeightGreaterThanOne;
		MIMATRIX m_Matrix;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_MULTIPLICATIVE (const FILTER_FOCAL_MULTIPLICATIVE&);
		FILTER_FOCAL_MULTIPLICATIVE& operator= (const FILTER_FOCAL_MULTIPLICATIVE&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to apply two kernels on samples in image.  Used for edge detection.
class FILTER_FOCAL_GRADIENT : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_GRADIENT (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			const MIMATRIX& matrixA,				//!< Set kernel A from matrix
			const MIMATRIX& matrixB,				//!< Set kernel B from matrix
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_GRADIENT (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			const MIMATRIX& matrixA,				//!< Set kernel A from matrix
			const MIMATRIX& matrixB,				//!< Set kernel B from matrix
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_GRADIENT ();

		//! Get matrix kernel A
		const MIMATRIX& GetMatrixA (
			) const { return m_MatrixA; }

		//! Use matrix to define kernel A
		void SetMatrixA (
			const MIMATRIX& matrix
			) { m_MatrixA = matrix; }

		//! Get matrix kernel B
		const MIMATRIX& GetMatrixB (
			) const { return m_MatrixB; }

		//! Use matrix to define kernel B
		void SetMatrixB (
			const MIMATRIX& matrix
			) { m_MatrixB = matrix; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MIMATRIX m_MatrixA;
		MIMATRIX m_MatrixB;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_GRADIENT (const FILTER_FOCAL_GRADIENT&);
		FILTER_FOCAL_GRADIENT& operator= (const FILTER_FOCAL_GRADIENT&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to compute statistics on window
class FILTER_FOCAL_STATISTICS : public FILTER_FOCAL {
	public:

		//! Statistic type
		enum STAT_TYPE {
			STAT_TYPE_Mean,
			STAT_TYPE_Median,
			STAT_TYPE_Majority,
			STAT_TYPE_Minority,
			STAT_TYPE_Maximum,
			STAT_TYPE_Minimum,
			STAT_TYPE_StdDev,
			STAT_TYPE_Sum,
			STAT_TYPE_Variance,
			STAT_TYPE_Variety
			};

		//! Constructor.
		FILTER_FOCAL_STATISTICS (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAT_TYPE Type,							//!< Statistic type to compute
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_STATISTICS (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			STAT_TYPE Type,							//!< Statistic type to compute
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_STATISTICS ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		STAT_TYPE m_Type;
		DOUBLE_ARRAY<double> m_ValueArray;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_STATISTICS (const FILTER_FOCAL_STATISTICS&);
		FILTER_FOCAL_STATISTICS& operator= (const FILTER_FOCAL_STATISTICS&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to apply LUM filter (Lower-Upper-Middle)
//! The Lower-Upper-Middle (LUM) filter is a nonlinear edge-enhancement filter that simultaneously suppresses image noise. 
//! The filter first ranks the values within the filter window in numerical order. 
//! Two user-defined parameters are used to partition the resulting list of values (number of values = N), and the filter output depends upon which partition includes the value in the center of the filter window (the filter input value). 
//! Edge Parameter is an integer value that provides upper and lower rank limits for the values near the median value. 
//! If the filter input value is within the range set by Parameter A (that is, near the local median), the filter outputs the lower limit value if the input is below the median, or the upper limit value if the input value is above the median
//! Noise Parameter controls the noise-suppression function of the filter by setting overall minimum and maximum values for the filter output.
class FILTER_FOCAL_LUM : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_LUM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			int EdgeParam,								//!< Edge Enhancement Parameter, provides upper and lower rank limits for the values near the median value
			int NoiseParam,							//!< Noise Supression Parameter, sets overall minimum and maximum values for the filter output
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_LUM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			int EdgeParam,								//!< Edge Enhancement Parameter, provides upper and lower rank limits for the values near the median value
			int NoiseParam,							//!< Noise Supression Parameter, sets overall minimum and maximum values for the filter output
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_LUM ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_EdgeParam;
		int m_NoiseParam;
		DOUBLE_ARRAY<double> m_ValueArray;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_LUM (const FILTER_FOCAL_LUM&);
		FILTER_FOCAL_LUM& operator= (const FILTER_FOCAL_LUM&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to apply CS filter (comparison and selection)
//! The filter enhances edges by converting raster values that are close to the local mean to new, locally defined values that are farther from the mean. 
//! The CS filter ranks the values within the filter window in numerical order (number of values = N), and calculates the mean raster value for the current filter window. 
//! User defined Rank Parameter identifies a pair of rank numbers symmetrical about the mean. 
//! The raster values corresponding to these rank numbers for the current filter window provide the two possible choices for the filter output value. 
//! If the filter input value is less than the mean, the lower raster value is used for the filter output. If the filter input value is greater than the mean, the higher value is used.
class FILTER_FOCAL_CS : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_CS (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			int RankParam,								//!< Rank Parameter, defines the two possible output values used from the sorted array
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_CS (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			int RankParam,								//!< Rank Parameter, defines the two possible output values used from the sorted array
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_CS ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_RankParam;
		DOUBLE_ARRAY<double> m_ValueArray;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_CS (const FILTER_FOCAL_CS&);
		FILTER_FOCAL_CS& operator= (const FILTER_FOCAL_CS&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Filter to apply WMMR-MED filter (weighted majority with minimum range - median)
//! The process first ranks the values within the filter window in numerical order (number of values = N). 
//! The filter then finds a fixed-size subset in which the values are most closely spaced, and calculates a median value from it (WMMR-MED value). 
//! The number of values in the subset is (N + 1) / 2. 
//! The output value of the filter is a user-controlled weighted average of the input raster value and the WMMR-MED value.
class FILTER_FOCAL_WMMR : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_WMMR (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the WMMR-MED value
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_WMMR (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the WMMR-MED value
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_WMMR ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_WeightParam;
		DOUBLE_ARRAY<double> m_ValueArray;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_WMMR (const FILTER_FOCAL_WMMR&);
		FILTER_FOCAL_WMMR& operator= (const FILTER_FOCAL_WMMR&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The P-Median filter is designed to suppress noise while preserving edge and line detail. 
//! The filter calculates median values for two subsets of the values in the filter window: 
//! 1) combined horizontal and vertical transects through the center cell, 
//! and 2) two diagonal transects through the center cell. These two median values are then averaged. 
//! The output of the filter is a user-controlled weighted average of the averaged median and the original center cell value.
class FILTER_FOCAL_PMEDIAN : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_PMEDIAN (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the P-Median value
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_PMEDIAN (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the P-Median value
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_PMEDIAN ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_WeightParam;
		DOUBLE_ARRAY<double> m_ValueArrayA;
		DOUBLE_ARRAY<double> m_ValueArrayB;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_PMEDIAN (const FILTER_FOCAL_PMEDIAN&);
		FILTER_FOCAL_PMEDIAN& operator= (const FILTER_FOCAL_PMEDIAN&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Lee filter uses a least-squares approach to estimate the true signal strength of the center cell in the filter window from the measured value in that cell, the local mean brightness of all cells in the window, and a gain factor calculated from the local variance and the noise standard deviation. 
//! The filter assumes a Gaussian (normal) distribution for the noise values, and calculates the local noise standard deviation for each filter window. 
//! The Lee filter calculation produces an output value close to the local mean for uniform areas, and a value close to the original input value in higher contrast regions. 
//! Most smoothing occurs in the more uniform areas, while edges and other fine detail is maintained.
class FILTER_FOCAL_LEE : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_LEE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_LEE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_LEE ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_LEE (const FILTER_FOCAL_LEE&);
		FILTER_FOCAL_LEE& operator= (const FILTER_FOCAL_LEE&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! Multi Level Median (MLM)
//! The MLM filter is designed to reduce image noise (outlier values) while preserving edges, corners, and thin line detail in the image. 
//! The filter calculates separate median values for horizontal, vertical, and two diagonal transects through the central cell in the filter window. 
//! The minimum and maximum of these four values are then found. 
//! The minimum, maximum, and original center raster value are then ranked, and the median of these three values is assigned as the filter output value.
class FILTER_FOCAL_MLM : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_MLM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_MLM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_MLM ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		DOUBLE_ARRAY<double> m_ValueArrayA;
		DOUBLE_ARRAY<double> m_ValueArrayB;
		DOUBLE_ARRAY<double> m_ValueArrayC;
		DOUBLE_ARRAY<double> m_ValueArrayD;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_MLM (const FILTER_FOCAL_MLM&);
		FILTER_FOCAL_MLM& operator= (const FILTER_FOCAL_MLM&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Range filter produces an image of one of the simplest elements of texture, the local range of values. 
//! The output of the range filter is the range of values in the filter window (difference between local minimum and maximum values) multiplied by an adjustable gain factor (to produce an image with sufficient brightness and contrast).
class FILTER_FOCAL_RANGE : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_RANGE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			double ContrastParam,					//!< Adjustable gain factor
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_RANGE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			double ContrastParam,					//!< Adjustable gain factor
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_RANGE ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_ContrastParam;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_RANGE (const FILTER_FOCAL_RANGE&);
		FILTER_FOCAL_RANGE& operator= (const FILTER_FOCAL_RANGE&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Olympic filter is a variant of a low pass (averaging) filter.
//! The Olympic filter ranks the values within the filter window (number of values = N), and discards high and low values before calculating the mean of those remaining.
class FILTER_FOCAL_OLYMPIC : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_OLYMPIC (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			int NumDiscard,							//!< Number of values discarded at each end of the distribution
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_OLYMPIC (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			int NumDiscard,							//!< Number of values discarded at each end of the distribution
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_OLYMPIC ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		int m_NumDiscard;
		DOUBLE_ARRAY<double> m_ValueArray;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_OLYMPIC (const FILTER_FOCAL_OLYMPIC&);
		FILTER_FOCAL_OLYMPIC& operator= (const FILTER_FOCAL_OLYMPIC&);
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Teager filter produces an image of the edges within the original scene. 
//! The Teager filter uses the same variable edge-enhancement technique as the Volterra / Unsharp filter process, but without adding the original input image value back in to produce the final output raster.
class FILTER_FOCAL_TEAGER : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_TEAGER (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_TEAGER (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_TEAGER ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MIMATRIX m_Matrix;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_TEAGER (const FILTER_FOCAL_TEAGER&);
		FILTER_FOCAL_TEAGER& operator= (const FILTER_FOCAL_TEAGER&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Volterra / Unsharp filter process uses a quadratic Volterra filter to enhance edges. 
//! The output of the Volterra filter is approximately equivalent to the product of the local mean and a high-pass filter. 
//! The result is then scaled and added to the original input image value to produce the final output.
class FILTER_FOCAL_VOLTERRA : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_VOLTERRA (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			double ContrastParam,					//!< Controls scaling of edge enhancement
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_VOLTERRA (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			double ContrastParam,					//!< Controls scaling of edge enhancement
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_VOLTERRA ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_ContrastParam;

		MIMATRIX m_Matrix;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_VOLTERRA (const FILTER_FOCAL_VOLTERRA&);
		FILTER_FOCAL_VOLTERRA& operator= (const FILTER_FOCAL_VOLTERRA&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Frost filter is an adaptive radar filter that incorporates the local image statistics in the filtering process, assuming a negative exponential distribution for the speckle noise. 
//! The filter performs a weighted average of the cell values in the filter window, with the weights for each cell being determined from the local statistics to minimize the mean square error of the signal estimate. 
//! The filter weight for a cell is a negative exponential function of the noise standard deviation (calculated locally for each filter window), and also decrease with distance from the center cell. 
//! The center cells are weighted more heavily as the variance in the filter window increases. 
//! The filter therefore smoothes more in homogeneous areas, but provides a signal estimate closer to the observed value of the center cell in heterogeneous areas.
class FILTER_FOCAL_FROST : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_FROST (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_FROST (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_FROST ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MIMATRIX m_MatDistance;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_FROST (const FILTER_FOCAL_FROST&);
		FILTER_FOCAL_FROST& operator= (const FILTER_FOCAL_FROST&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Adaptive Mean P-Median filter (AMPM) is a variant of the P-Median filter that is designed to provide better smoothing in uniform regions while still preserving edges and line detail. 
//! The filter first classifies the center cell in the filter window as belonging to either a uniform region or an edge region. 
//! The AMPM filter then applies the P-Median filter in regions containing edges, but uses a simple averaging filter in regions with relatively uniform values. 
//! The output of the AMPM filter in the edge regions is a user-controlled weighted average of the calculated P-Median filter value and the original center cell value; this weighting does not apply to the areas where simple averaging is performed.
class FILTER_FOCAL_AMPM : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_AMPM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the P-Median value
			double SigmaParam,						//!< User-controlled weight applied to sigma
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_AMPM (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double WeightParam,						//!< User-controlled weighted average of the input raster value and the P-Median value
			double SigmaParam,						//!< User-controlled weight applied to sigma
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns							//!< Number of columns in processing window
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_AMPM ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_WeightParam;
		double m_SigmaParam;
		DOUBLE_ARRAY<double> m_Sigmas;
		STATSVALIDITY m_StatsValidity;
		DOUBLE_ARRAY<double> m_ValueArrayA;
		DOUBLE_ARRAY<double> m_ValueArrayB;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_AMPM (const FILTER_FOCAL_AMPM&);
		FILTER_FOCAL_AMPM& operator= (const FILTER_FOCAL_AMPM&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Kuan Adaptive Noise Smoothing filter uses a minimum mean square error calculation to estimate the value of the true signal for the center cell in the filter window from the local statistics. 
//! It is similar in approach to the Lee filter, but makes fewer simplifying assumptions in the calculations. 
//! The Adaptive Noise Filter calculates the signal estimate from the local mean and variance, and the noise standard deviation (assumed to be constant for the entire image); it assumes a Gaussian (normal) distribution for the speckle noise.
class FILTER_FOCAL_KUANADAPTIVE : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_KUANADAPTIVE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,						//!< User-controlled weight applied to sigma
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_KUANADAPTIVE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,						//!< User-controlled weight applied to sigma
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_KUANADAPTIVE ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_SigmaParam;
		DOUBLE_ARRAY<double> m_Sigmas;
		STATSVALIDITY m_StatsValidity;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_KUANADAPTIVE (const FILTER_FOCAL_KUANADAPTIVE&);
		FILTER_FOCAL_KUANADAPTIVE& operator= (const FILTER_FOCAL_KUANADAPTIVE&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Sigma filter assumes that for a particular average image brightness, the variation in brightness due to speckle noise has a Gaussian (normal) distribution. 
//! For each filter window, the filter output is the average of the cells whose values fall within a prescribed range above and below that of the center cell value. 
//! The limits of this range are set as a user-defined multiple of the noise standard deviation. 
//! Variation within these limits is assumed to be the result of speckle noise in a uniform area, whereas cells with values outside these limits probably represent different surface materials, and thus should not contribute to the averaging process.
class FILTER_FOCAL_SIGMA : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_SIGMA (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,						//!< User-controlled weight applied to sigma
			int KThreshold,							//!< If the number of cells within the standard deviation limits is less than threshold, center value is replaced by a simple average of the surrounding cells
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_SIGMA (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnershipSource,				//!< Filter takes ownership of source image
			STAGE *pMaskImage,						//!< Mask, must have single component sample
			bool TakeOwnershipMask,					//!< Filter takes ownership of mask stage and will delete when needed
			STATSVALIDITY StatsValidity,			//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,						//!< User-controlled weight applied to sigma
			int KThreshold,							//!< If the number of cells within the standard deviation limits is less than threshold, center value is replaced by a simple average of the surrounding cells
			UINT32 NumFocalRows,						//!< Number of rows in processing window
			UINT32 NumFocalColumns,					//!< Number of columns in processing window
			double AmountFilterPct = 100.0		//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_SIGMA ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_SigmaParam;
		DOUBLE_ARRAY<double> m_Sigmas;
		STATSVALIDITY m_StatsValidity;
		int m_KThreshold;
		DOUBLE_ARRAY<double> m_ValueArrayA;
		DOUBLE_ARRAY<double> m_ValueArrayB;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_SIGMA (const FILTER_FOCAL_SIGMA&);
		FILTER_FOCAL_SIGMA& operator= (const FILTER_FOCAL_SIGMA&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

//! The Local Contrast filter performs a locally adaptive, spatially varying contrast enhancement. 
//! The amount of enhancement varies as a function of the existing contrast in the local neighborhood, expressed as the ratio of the global mean to the local standard deviation. 
//! Areas with low inherent contrast are enhanced more than areas which already have high contrast
class FILTER_FOCAL_CONTRAST : public FILTER_FOCAL {
	public:

		//! Constructor.
		FILTER_FOCAL_CONTRAST (
			STAGE *pSourceImage,							//!< Source image stage
			bool TakeOwnershipSource,					//!< Filter takes ownership of source image
			STATSVALIDITY StatsValidity,				//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,							//!< User-controlled weight applied to sigma
			double ContrastParam,						//!< Amount of local contrast enhancement that is applied throughout the image
			UINT32 NumFocalRows,							//!< Number of rows in processing window
			UINT32 NumFocalColumns,						//!< Number of columns in processing window
			double AmountFilterPct = 100.0			//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Constructor with mask to determine areas to filter.
		FILTER_FOCAL_CONTRAST (
			STAGE *pSourceImage,							//!< Source image stage
			bool TakeOwnershipSource,					//!< Filter takes ownership of source image
			STAGE *pMaskImage,							//!< Mask, must have single component sample
			bool TakeOwnershipMask,						//!< Filter takes ownership of mask stage and will delete when needed
			STATSVALIDITY StatsValidity,				//!< How to compute image statisitcs used in filter (fastest should be used for display, exact for processing)
			double SigmaParam,							//!< User-controlled weight applied to sigma
			double ContrastParam,						//!< Amount of local contrast enhancement that is applied throughout the image			
			UINT32 NumFocalRows,							//!< Number of rows in processing window
			UINT32 NumFocalColumns,						//!< Number of columns in processing window
			double AmountFilterPct = 100.0			//!< Amount of filter to apply (percentage, 0-100)
			);

		//! Destructor.
		virtual ~FILTER_FOCAL_CONTRAST ();		

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_ContrastParam;
		double m_SigmaParam;

		DOUBLE_ARRAY<double> m_Means;
		DOUBLE_ARRAY<double> m_Sigmas;
		STATSVALIDITY m_StatsValidity;
		DOUBLE_ARRAY<double> m_ValueArrayA;
		DOUBLE_ARRAY<double> m_ValueArrayB;

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_FOCAL_CONTRAST (const FILTER_FOCAL_CONTRAST&);
		FILTER_FOCAL_CONTRAST& operator= (const FILTER_FOCAL_CONTRAST&);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERFOCAL_H
