/**
 * \file spatmod/imagestageautoptr.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::STAGE_AUTOPTR definition.
 *
 * \if NODOC
 *
 * $Id: imagestageautoptr.h_v 1.4 2012/01/26 16:37:27 mju Exp $
 *
 * $Log: imagestageautoptr.h_v $
 * Revision 1.4  2012/01/26 16:37:27  mju
 * Change detach to reset stage ptr to 0 and not change ownership.
 *
 * Revision 1.3  2006/10/16 17:42:59  mju
 * Add detach method.
 *
 * Revision 1.2  2006/10/13 14:41:18  mju
 * Add isAssigned.
 *
 * Revision 1.1  2006/10/11 14:10:03  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGESTAGEAUTOPTR_H
#define	INC_SPATMOD_IMAGESTAGEAUTOPTR_H

namespace SPATMOD {
namespace IMAGE {

class STAGE;

//=====================================================================================================================

//! Automatic STAGE pointer which keeps track of ownership and destroys when needed.
class STAGE_AUTOPTR {
	public:

		//! Default constructor.
		STAGE_AUTOPTR (
			): m_pStage(0), m_IsOwner(false) { }

		//! Construct and attach stage.
		STAGE_AUTOPTR (
			STAGE *pStage,
			bool TakeOwnership
			): m_pStage(pStage), m_IsOwner(TakeOwnership) { }

		//! Copy constructor, copy will not have ownership.
		STAGE_AUTOPTR (
			const STAGE_AUTOPTR& rhs
			): m_pStage(rhs.m_pStage), m_IsOwner(false) { }

		//! Destructor.
		~STAGE_AUTOPTR ();

		//! Assignment operator, does not transfer ownership.
		STAGE_AUTOPTR& operator= (
			const STAGE_AUTOPTR& rhs
			);

		//! Dereference operator.
		STAGE& operator* (
			) const { return (*m_pStage); }

		//! Pointer operator.
		STAGE* operator-> (
			) const { return (m_pStage); }

		//! Cast to STAGE*.
		operator STAGE* (
			) const { return (m_pStage); }

		//! Assign with specified ownership.
		void Assign (
			STAGE *pStage,
			bool TakeOwnership
			);

		//! Assign with specified ownership.
		void Assign (
			STAGE_AUTOPTR& StagePtr,
			bool TakeOwnership
			);

		//! Detach associated stage.
		//! @return Associated stage, caller must ensure deletion if STAGE_AUTO_PTR was previous owner.
		STAGE* Detach (
			) { STAGE *pStage = m_pStage; m_pStage = 0; return (pStage); }

		//! Determine if has stage assigned.
		bool IsAssigned (
			) const { return (m_pStage != 0); }

		//! Determine if owner of stage pointer.
		bool IsOwner (
			) const { return (m_IsOwner); }

   	#ifndef GENERATING_DOXYGEN_OUTPUT
		//! Set whether is owner of pointer or not.
		//! Use with caution, improper use will result in memory leak or crash.
		//! Intended for use internal to pipeline only.
		void SetOwner (
			bool TakeOwnership
			) { m_IsOwner = TakeOwnership; }
		#endif // GENERATING_DOXYGEN_OUTPUT

	private:
   	#ifndef GENERATING_DOXYGEN_OUTPUT
		STAGE *m_pStage;
		bool m_IsOwner;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGESTAGEAUTOPTR_H
